@@ -14,6 +14,7 @@
 #if defined(NeXT)
 #undef  HAVE_SYS_UTSNAME_H
 #define HAVE_DYN_UNDERSTORE 1
+#undef	HAVE_SBRK
 #endif
 
 /* Hack to configure AmigaOS */
@@ -26,6 +27,19 @@
 #undef	__WIN32__
 #define __WIN32__ 1
 #endif
+#if defined(__WIN32__)
+#if defined(HAVE_WINSOCK2_H)
+#include <winsock2.h>
+#if _MSC_VER < 1200
+#define HAVE_INT32 1
+#endif
+#endif
+#if defined(__cplusplus)
+#define HAVE_BOOL 1
+#else
+#define inline
+#endif
+#endif
 
 /* Hack for NetBSD */
 #if defined(__NetBSD__)
@@ -38,9 +52,15 @@
 #define  HAVE_SYS_SELECT_H 1
 #endif
 
-/* Hack to configure Linux with WINE installed */
-#if defined(linux)
-#undef	HAVE_WINSOCK_H
+/* Hack to configure BeOS R4 */
+#if defined(__BEOS__)
+#include <socket.h>
+#if defined(MSG_PEEK)
+#error Looks like BeOS finally supports MSG_PEEK.
+#error Please remove the MSG_PEEK hack in config-hacks.h and syscalls.c.
+#else
+#define MSG_PEEK 0x2
+#endif
 #endif
 
 /* The 386 can do unaligned memory accesses */
@@ -52,7 +72,10 @@
  * word order for its doubles.  The arm32 uses software floating point 
  * emulation.  We define this constant to declare that the word order
  * in a 64bit double is the opposite of the word order in a 64bit int.
+ *
+ * If you change this definition, also change strtod("-0.0") check
+ * in configure.in.
  */
-#if defined(arm32)
+#if defined(arm32) || (defined(arm) && defined(linux))
 #define DOUBLE_ORDER_OPPOSITE
 #endif
