@@ -36,6 +36,14 @@
 #include <sigcontext.h>
 #endif
 
+/*
+ * newer Linux kernel actually implement SA_SIGINFO.
+ * But we don't need it, so let's turn it off
+ */
+#if defined(SA_SIGINFO)
+#undef SA_SIGINFO
+#endif
+
 /* Function prototype for signal handlers */
 #if defined(HAVE_STRUCT_SIGCONTEXT_STRUCT) && !defined(__GLIBC__)
 /* Linux < 2.1.1 */
@@ -46,6 +54,8 @@
 /* Linux >= 2.1.1  or Linux 2.0.x with glibc2 */
 #define	EXCEPTIONPROTO							\
 	int sig, struct sigcontext ctx
+#else
+#error Do not know how to define EXCEPTIONPROTO
 #endif
 
 /* Get the first exception frame from a signal handler */
