@@ -14,30 +14,27 @@
 
 #include "object.h"
 #include "native.h"
-#include "java_lang_String.h"
 #include "java_lang_Throwable.h"
 
-#define STRING_SIZE(STR) \
-		javaStringLength(STR)
-#define STRING_DATA(STR) \
-		&(unhand(unhand(STR)->value)->body[unhand(STR)->offset])
-
 #define	STRINGCLASS	"java/lang/String"
-#define	STRINGINIT	"<init>"
-#define	STRINGINITSIG	"([C)V"
-
 #define	OBJECTCLASS	"java/lang/Object"
 #define	CLASSCLASS	"java/lang/Class"
 
 #define	OBJECTCLASSSIG	"Ljava/lang/Object;"
 
+#define	PTRCLASS	"kaffe/util/Ptr"
+#define	PTRCLASSSIG	"Lkaffe/util/Ptr;"
+
 void initBaseClasses(void);
+void initialiseKaffe(void);
 
 extern struct Hjava_lang_Class*	ClassClass;
 extern struct Hjava_lang_Class*	ObjectClass;
 extern struct Hjava_lang_Class*	StringClass;
 extern struct Hjava_lang_Class*	SystemClass;
 extern struct Hjava_lang_Class*	SerialClass;
+extern struct Hjava_lang_Class*	CloneClass;
+extern struct Hjava_lang_Class* PtrClass;
 
 extern struct Hjava_lang_Class*	javaLangVoidClass;
 extern struct Hjava_lang_Class*	javaLangBooleanClass;
@@ -49,4 +46,8 @@ extern struct Hjava_lang_Class*	javaLangLongClass;
 extern struct Hjava_lang_Class*	javaLangFloatClass;
 extern struct Hjava_lang_Class*	javaLangDoubleClass;
 
+extern struct Hjava_lang_Class*	javaLangNullPointerException;
+extern struct Hjava_lang_Class*	javaLangArithmeticException;
+extern struct Hjava_lang_Class* javaLangArrayIndexOutOfBoundsException;
+
 #endif
