@@ -13,20 +13,33 @@
 
 typedef	struct {
 	Hjava_lang_Class*	type;
+	uint8			used;
+	uint8			modified;
 } frameElement;
 
 typedef struct perPCInfo {
 	uint16			stackPointer;
 	uint16			flags;
+#if defined(TRANSLATOR)
 	uintp			nativePC;
+#endif
 	struct perPCInfo*	nextBB;
 	frameElement*		frame;
 } perPCInfo;
 
+typedef struct _localUse {
+	int32			use;
+	int32			first;
+	int32			last;
+	int32			write;
+	Hjava_lang_Class*	type;
+} localUse;
+
 typedef struct codeinfo {
 	uint16			codelen;
 	uint16			stacksz;
 	uint16			localsz;
+	localUse*		localuse;
 	perPCInfo		perPC[1];
 } codeinfo;
 
@@ -34,11 +47,11 @@ typedef struct codeinfo {
 #define	TUNSTABLE		((Hjava_lang_Class*)1)
 #define	TADDR			((Hjava_lang_Class*)2)
 #define	TOBJ			((Hjava_lang_Class*)3)
-#define	TVOID			(voidClass)
-#define	TINT			(intClass)
-#define	TLONG			(longClass)
-#define	TFLOAT			(floatClass)
-#define	TDOUBLE			(doubleClass)
+#define	TVOID			(_Jv_voidClass)
+#define	TINT			(_Jv_intClass)
+#define	TLONG			(_Jv_longClass)
+#define	TFLOAT			(_Jv_floatClass)
+#define	TDOUBLE			(_Jv_doubleClass)
 
 #define	CONSTANTTYPE(VAL, LCL)					\
 	switch (CLASS_CONST_TAG(meth->class, (LCL))) {		\
@@ -67,7 +80,7 @@ typedef struct codeinfo {
 #define	FLAG_STARTOFEXCEPTION		0x0002
 #define	FLAG_STACKPOINTERSET		0x0004
 #define	FLAG_NORMALFLOW			0x0008
-#define	FLAG_FLOW			0x0010
+#define	FLAG_JUMPFLOW			0x0010
 #define	FLAG_JUMP			0x0020
 #define	FLAG_NEEDVERIFY			0x0040
 #define	FLAG_DONEVERIFY			0x0080
@@ -93,11 +106,23 @@ typedef struct codeinfo {
 					SET_NEWFRAME(PC); \
 					FLAGS(PC) |= FLAG_STARTOFBASICBLOCK
 
-#define	SET_STACKPOINTER(PC, SP)	STACKPOINTER(PC) = (SP); \
-					FLAGS(PC) |= FLAG_STACKPOINTERSET
+#define	SET_STACKPOINTER(PC, SP)	do { \
+	if ((FLAGS(PC) & FLAG_STACKPOINTERSET) && STACKPOINTER(PC) != (SP)) { \
+		failed = true; \
+		DBG(CODEANALYSE, \
+			dprintf("sp size change at pc %d: %d != %d\n", (PC), STACKPOINTER(PC), (SP)); \
+		   ) \
+	} \
+	STACKPOINTER(PC) = (SP); \
+	FLAGS(PC) |= FLAG_STACKPOINTERSET; \
+} while(0)
 
+#if defined(TRANSLATOR)
 #define	SET_INSNPC(P, V)		codeInfo->perPC[(P)].nativePC = (V)
 #define	INSNPC(P)			codeInfo->perPC[(P)].nativePC
+#else
+#define	SET_INSNPC(P, V)
+#endif
 
 #if defined(INTERPRETER)
 #define	SET_INSN(PC, V)			INSN(PC) = (V)
@@ -105,7 +130,7 @@ typedef struct codeinfo {
 #define	SET_INSN(PC, V)
 #endif
 #define	SET_NORMALFLOW(pc)		FLAGS(pc) |= FLAG_NORMALFLOW
-#define	SET_JUMPFLOW(from, to)		FLAGS(to) |= FLAG_FLOW; \
+#define	SET_JUMPFLOW(from, to)		FLAGS(to) |= FLAG_JUMPFLOW; \
 					FLAGS(from) |= FLAG_JUMP
 #define	SET_STARTOFEXCEPTION(pc)	FLAGS(pc) |= FLAG_STARTOFEXCEPTION
 #define	SET_NEEDVERIFY(pc)		FLAGS(pc) |= FLAG_NEEDVERIFY
@@ -120,11 +145,15 @@ typedef struct codeinfo {
 #define	IS_NEEDVERIFY(pc)		(FLAGS(pc) & FLAG_NEEDVERIFY)
 #define	IS_DONEVERIFY(pc)		(FLAGS(pc) & FLAG_DONEVERIFY)
 #define	IS_STARTOFINSTRUCTION(pc)	(FLAGS(pc) & FLAG_STARTOFINSTRUCTION)
+#define	IS_UNREACHABLE(pc)		((IS_STARTOFBASICBLOCK(pc) || \
+					  IS_STARTOFEXCEPTION(pc)) && \
+					  !IS_DONEVERIFY(pc))
 
-#define	ALLOCFRAME()			gc_malloc_fixed((codeInfo->stacksz+codeInfo->localsz+1) * sizeof(frameElement))
+#define	ALLOCFRAME()			KMALLOC((codeInfo->stacksz+codeInfo->localsz+1) * sizeof(frameElement))
 
 #define	ATTACH_NEW_BASICBLOCK(DPC)				\
-	if ((DPC) != 0 && !IS_STARTOFBASICBLOCK(DPC)) {		\
+	if ((DPC) != 0 && !IS_STARTOFBASICBLOCK(DPC) &&		\
+			  !IS_STARTOFEXCEPTION(DPC)) {		\
 		btail->nextBB = &codeInfo->perPC[DPC];		\
 		btail = btail->nextBB;				\
 	}
@@ -132,18 +161,22 @@ typedef struct codeinfo {
 #define	SET_NEWFRAME(pc)					\
 	if (FRAME(pc) == 0) {					\
 		FRAME(pc) = ALLOCFRAME();			\
-		assert(FRAME(pc) != 0);				\
+		if (!FRAME(pc)) {				\
+			meth->accflags &= ~ACC_VERIFIED;	\
+			tidyVerifyMethod(&codeInfo);		\
+			return false;				\
+		}						\
 	}
 
 #define	FRAMEMERGE(PC, SP)					\
 	SET_STACKPOINTER(PC, SP);				\
-	mergeFrame(PC, SP, activeFrame, meth);			\
+	mergeFrame(codeInfo, PC, SP, activeFrame, meth);	\
         if (!IS_DONEVERIFY(PC)) {				\
                 SET_NEEDVERIFY(PC);				\
         }
 
 #define	FRAMEMERGE_LOCALS(PC)					\
-	mergeFrame(PC, meth->localsz+meth->stacksz, activeFrame, meth); \
+	mergeFrame(codeInfo, PC, meth->localsz+meth->stacksz, activeFrame, meth); \
         if (!IS_DONEVERIFY(PC)) {				\
                 SET_NEEDVERIFY(PC);				\
         }
@@ -194,12 +227,17 @@ typedef struct codeinfo {
 					STACK_CHECKTYPE(S, T)
 #define	STACKOUT(S, T)			SF(S).type = (T)
 #define	STACKOUT_CONST(S, T, V)		SF(S).type = (T)
-#define	STACKOUT_LOCAL(S, T, L)		SF(S) = LCL(L)
-
-#define	LOCALOUT(L, T)			LCL(L).type = (T)
-#define LOCALOUT_STACK(L, T, S)		LCL(L) = SF(S)
-#define	LOCALINOUT(L, T)
+#define	STACKOUT_LOCAL(S, T, L)		SF(S) = LCL(L); \
+					LCL(L).used = 1
 
+#define	LOCALOUT(L, T)			LCL(L).type = (T); \
+					LCL(L).used = 1; \
+					LCL(L).modified = 1
+#define LOCALOUT_STACK(L, T, S)		LCL(L) = SF(S); \
+					LCL(L).used = 1; \
+					LCL(L).modified = 1
+#define	LOCALINOUT(L, T)		LCL(L).used = 1; \
+					LCL(L).modified = 1
 #define	STACKCOPY(F, T)			SF(T) = SF(F)
 #define	STACKOUT_STACK(T, X, F)		STACKCOPY(F, T)
 #define	STACKSWAP(F, T)			{				\
@@ -209,14 +247,15 @@ typedef struct codeinfo {
 						SF(F) = tmp;		\
 					}
 
-#define	LOCALINIT(L, T);		FRAME(pc)[(L)].type = (T)
-#define	LOCALIN(L, T)			LOCAL_CHECKTYPE(L, T);
+#define	LOCALINIT(L, T);		FRAME(pc)[(L)].type = (T); \
+					FRAME(pc)[(L)].used = 1
+#define	LOCALIN(L, T)			LOCAL_CHECKTYPE(L, T); \
+					LCL(L).used = 1
 
 struct _methods;
-void verifyMethod(struct _methods*);
-void tidyVerifyMethod(void);
+bool verifyMethod(struct _methods*, codeinfo**, errorInfo*);
+void tidyVerifyMethod(codeinfo**);
 
-extern codeinfo* codeInfo;
 extern const uint8 insnLen[];
 
 #endif
