@@ -15,7 +15,6 @@
 #include "config-mem.h"
 #include "jtypes.h"
 #include "errors.h"
-#include "slots.h"
 #include "access.h"
 #include "object.h"
 #include "constants.h"
@@ -24,17 +23,25 @@
 #include "exception.h"
 #include "thread.h"
 #include "baseClasses.h"
+#include "stringSupport.h"
 #include "machine.h"
 #include "locks.h"
+#include "soft.h"
 
 static void throwAbstractMethodError(void);
-static void throwNoSuchMethodError(void);
 
 /*
  * Lookup a method reference and get the various interesting bits.
+ *
+ * Return false if unsuccessful because of a malformed class file
+ *	or if the class couldn't be found or processed.
+ * Returns true otherwise. 
+ *
+ * Note that even if it returns true, the method may not be found.
+ * call->method is set to NULL in this case.
  */
-void
-getMethodSignatureClass(constIndex idx, Hjava_lang_Class* this, bool loadClass, callInfo* call)
+bool
+getMethodSignatureClass(constIndex idx, Hjava_lang_Class* this, bool loadClass, int isSpecial, callInfo* call, errorInfo *einfo)
 {
 	constants* pool;
 	constIndex ci;
@@ -42,12 +49,23 @@ getMethodSignatureClass(constIndex idx, Hjava_lang_Class* this, bool loadClass,
 	Hjava_lang_Class* class;
 	Utf8Const* name;
 	Utf8Const* sig;
+	Method* mptr;
+	int i;
+
+	call->class = 0;
+	call->method = 0;
+	call->signature = 0;
+	call->name = 0;
+	call->cname = 0;
 
 	pool = CLASS_CONSTANTS(this);
 	if (pool->tags[idx] != CONSTANT_Methodref &&
 	    pool->tags[idx] != CONSTANT_InterfaceMethodref) {
-DBG(MLOOKUP,	dprintf("No Methodref found\n");			)
-                throwNoSuchMethodError();
+DBG(RESERROR,	dprintf("No Methodref found for idx=%d\n", idx);	)
+		/* shouldn't that be ClassFormatError or something? */
+		postExceptionMessage(einfo, JAVA_LANG(NoSuchMethodError),
+			"method name unknown, tag = %d", pool->tags[idx]);
+                return (false);
 	}
 
 	ni = METHODREF_NAMEANDTYPE(idx, pool);
@@ -57,44 +75,76 @@ DBG(MLOOKUP,	dprintf("No Methodref found\n");			)
 	call->name = name;
 	call->signature = sig;
 
-	if (loadClass == false) {
-		call->class = 0;
-		call->method = 0;
-	}
-	else {
+	if (loadClass == true) {
 		ci = METHODREF_CLASS(idx, pool);
-		class = getClass(ci, this);
-		processClass(class, CSTATE_LINKED);
+		class = getClass(ci, this, einfo);
+		if (class == NULL) {
+			call->cname = WORD2UTF(pool->data[ci]);
+			countInsAndOuts(sig->data, &call->in, &call->out, &call->rettype);
+			return (false);
+		}
+		assert(class->state >= CSTATE_LINKED);
 
-DBG(MLOOKUP,	dprintf("getMethodSignatureClass(%s,%s,%s)\n",
-			class->name->data, name->data, sig->data);	)
+                if (isSpecial == 1) {
+                        if (!utf8ConstEqual(name, constructor_name) && class !=
+ this && instanceof(class, this)) {
+                                class = this->superclass;
+                        }
+                }
 
 		call->class = class;
+		call->cname = class->name;
 		call->method = 0;
 		/* Find method - we don't use findMethod(...) yet since this
 		 * will initialise our class (and we don't want to do that).
 		 */
+		mptr = 0;
 		for (; class != 0; class = class->superclass) {
-			Method* mptr = findMethodLocal(class, name, sig);
+			mptr = findMethodLocal(class, name, sig);
 			if (mptr != NULL) {
 				call->method = mptr;
 				break;
 			}
 		}
 
+                /* If we've not found anything and we're searching interfaces,
+                 * search them too.
+                 */
+                if (mptr == 0 && isSpecial == 2) {
+                        class = call->class;
+                        for (i = class->total_interface_len - 1; i >= 0; i--) {
+                                mptr = findMethodLocal(class->interfaces[i], name, sig);
+                                if (mptr != 0) {
+                                        call->method = mptr;
+                                        break;
+                                }
+                        }
+                }
 	}
 
 	/* Calculate in's and out's */
 	countInsAndOuts(sig->data, &call->in, &call->out, &call->rettype);
+
+DBG(MLOOKUP,	
+	if (loadClass) 
+		dprintf("getMethodSignatureClass(%s,%s,%s) -> %s\n",
+			call->class->name->data, name->data, sig->data, 
+			(call->method ? "success" : "failure"));	)
+	return (true);
 }
 
+/*
+ * Get and link the class to which constant pool index idx in class this
+ * refers.  The returned class object is at least LINKED.
+ */
 Hjava_lang_Class*
-getClass(constIndex idx, Hjava_lang_Class* this)
+getClass(constIndex idx, Hjava_lang_Class* this, errorInfo *einfo)
 {
 	constants* pool;
 	Utf8Const *name;
 	Hjava_lang_Class* class;
 	int tag;
+	int iLockRoot;
 
 	pool = CLASS_CONSTANTS(this);
 
@@ -120,21 +170,21 @@ getClass(constIndex idx, Hjava_lang_Class* this)
 		break;
 
 	default:
-		throwException(ClassFormatError);
-		break;
+		postException(einfo, JAVA_LANG(ClassFormatError));
+		return NULL;
 	}
 
-	/* Find the specified class.  We cannot use 'loadClassOrArray' here
-	 * because the name is *not* a signature.
+	/* Find the specified class.
+	 * NB: the name is *not* a signature.
 	 */
 	if (name->data[0] == '[') {
-		class = loadArray(name, this->loader);
+		class = loadArray(name, this->loader, einfo);
 	}
 	else {
-		class = loadClass(name, this->loader);
+		class = loadClass(name, this->loader, einfo);
 	}
-	if (class == 0) {
-		throwException(ClassNotFoundException(name->data));
+	if (class == 0) {	/* propagate failure */
+		return NULL;
 	}
 
 	/* Lock the class while we update the constant pool.  Someone
@@ -148,8 +198,8 @@ getClass(constIndex idx, Hjava_lang_Class* this)
 	return (class);
 }
 
-void
-getField(constIndex idx, Hjava_lang_Class* this, bool isStatic, fieldInfo* ret)
+bool
+getField(constIndex idx, Hjava_lang_Class* this, bool isStatic, fieldInfo* ret, errorInfo *einfo)
 {
 	constants* pool;
 	constIndex ci;
@@ -159,29 +209,39 @@ getField(constIndex idx, Hjava_lang_Class* this, bool isStatic, fieldInfo* ret)
 
 	pool = CLASS_CONSTANTS(this);
 	if (pool->tags[idx] != CONSTANT_Fieldref) {
-DBG(FLOOKUP,	dprintf("No Fieldref found\n");				)
-                throwException(NoSuchFieldError(""));
+DBG(RESERROR,	dprintf("No Fieldref found\n");				)
+		postExceptionMessage(einfo, JAVA_LANG(NoSuchFieldError),
+			"tag was %d", pool->tags[idx]);
+		return (false);
 	}
 
 	ci = FIELDREF_CLASS(idx, pool);
-	class = getClass(ci, this);
+
+	class = getClass(ci, this, einfo);
+	if (class == NULL) {
+		return (false);
+	}
 
 	ni = FIELDREF_NAMEANDTYPE(idx, pool);
 
+	ret->cname = class->name;
+	ret->name = WORD2UTF(pool->data[NAMEANDTYPE_NAME(ni, pool)]);
+	ret->signature = WORD2UTF(pool->data[NAMEANDTYPE_SIGNATURE(ni, pool)]);
+
 DBG(FLOOKUP,	dprintf("*** getField(%s,%s,%s)\n",
 		class->name->data, 
 		WORD2UTF(pool->data[NAMEANDTYPE_NAME(ni, pool)])->data, 
 		WORD2UTF(pool->data[NAMEANDTYPE_SIGNATURE(ni, pool)])->data);
     )
 
-	field = lookupClassField(class, WORD2UTF(pool->data[NAMEANDTYPE_NAME(ni, pool)]), isStatic);
+	field = lookupClassField(class, WORD2UTF(pool->data[NAMEANDTYPE_NAME(ni, pool)]), isStatic, einfo);
 	if (field == 0) {
-DBG(FLOOKUP,	printf("Field not found\n");				)
-                throwException(NoSuchFieldError(WORD2UTF(pool->data[NAMEANDTYPE_NAME(ni, pool)])->data));
+		return (false);
 	}
 
 	ret->field = field;
 	ret->class = class;
+	return (true);
 }
 
 /*
@@ -192,34 +252,49 @@ findMethodLocal(Hjava_lang_Class* class, Utf8Const* name, Utf8Const* signature)
 {
 	Method* mptr;
 	int n;
+
 	/*
 	 * Lookup method - this could be alot more efficient but never mind.
 	 * Also there is no attempt to honour PUBLIC, PRIVATE, etc.
 	 */
 	n = CLASS_NMETHODS(class);
 	for (mptr = CLASS_METHODS(class); --n >= 0; ++mptr) {
-		if (equalUtf8Consts (name, mptr->name) && equalUtf8Consts (signature, mptr->signature)) {
+		if (utf8ConstEqual (name, mptr->name) && utf8ConstEqual (signature, METHOD_SIG(mptr))) {
 			if ((mptr->accflags & ACC_ABSTRACT) != 0) {
 				SET_METHOD_NATIVECODE(mptr, (void*)throwAbstractMethodError);
 				mptr->accflags |= ACC_NATIVE;
 			}
+DBG(MLOOKUP,
+			dprintf("findMethodLocal(%s,%s,%s) -> %p\n",
+				class->name->data, name->data, signature->data, mptr); )
+
 			return (mptr);
 		}
 	}
-	return (NULL);
+
+DBG(MLOOKUP,
+	dprintf("findMethodLocal(%s,%s,%s) -> NOT FOUND\n",
+		class->name->data, name->data, signature->data, mptr); )
+
+	return NULL;
 }
 
 /*
  * Lookup a method (and translate) in the specified class.
  */
 Method*
-findMethod(Hjava_lang_Class* class, Utf8Const* name, Utf8Const* signature)
+findMethod(Hjava_lang_Class* class, Utf8Const* name, Utf8Const* signature, errorInfo *einfo)
 {
+	bool success;
 	/*
 	 * Waz CSTATE_LINKED - Must resolve constants before we do any
 	 * translation.  Might not be right though ... XXX
 	 */
-	processClass(class, CSTATE_OK);
+	if (class->state < CSTATE_USABLE) {
+		success = processClass(class, CSTATE_COMPLETE, einfo);
+		if (!success)
+			return (0);
+	}
 
 	/*
 	 * Lookup method - this could be alot more efficient but never mind.
@@ -231,92 +306,10 @@ findMethod(Hjava_lang_Class* class, Utf8Const* name, Utf8Const* signature)
 			return mptr;
 		}
 	}
+	postExceptionMessage(einfo, JAVA_LANG(NoSuchMethodError), name->data);
 	return (0);
 }
 
-#if defined(TRANSLATOR)
-/*
- * Find exception in method.
- */
-void
-findExceptionInMethod(uintp pc, Hjava_lang_Class* class, exceptionInfo* info)
-{
-	Method* ptr;
-
-	info->handler = 0;
-	info->class = 0;
-	info->method = 0;
-
-	ptr = findMethodFromPC(pc);
-	if (ptr != 0) {
-		if (findExceptionBlockInMethod(pc, class, ptr, info) == true) {
-			return;
-		}
-	}
-DBG(ELOOKUP,	dprintf("Exception not found.\n");			)
-}
-#endif
-
-/*
- * Look for exception block in method.
- */
-bool
-findExceptionBlockInMethod(uintp pc, Hjava_lang_Class* class, Method* ptr, exceptionInfo* info)
-{
-	jexceptionEntry* eptr;
-	Hjava_lang_Class* cptr;
-	int i;
-
-	/* Stash method & class */
-	info->method = ptr;
-	info->class = ptr->class;
-
-	eptr = &ptr->exception_table->entry[0];
-
-DBG(ELOOKUP,	
-	dprintf("Nr of exceptions = %d\n", ptr->exception_table->length); )
-
-	/* Right method - look for exception */
-	if (ptr->exception_table == 0) {
-		return (false);
-	}
-	for (i = 0; i < ptr->exception_table->length; i++) {
-		uintp start_pc = eptr[i].start_pc;
-		uintp end_pc = eptr[i].end_pc;
-		uintp handler_pc = eptr[i].handler_pc;
-
-DBG(ELOOKUP,	dprintf("Exceptions %x (%x-%x)\n", pc, start_pc, end_pc); )
-		if (pc < start_pc || pc > end_pc) {
-			continue;
-		}
-DBG(ELOOKUP,	dprintf("Found exception 0x%x\n", handler_pc); )
-
-		/* Found exception - is it right type */
-		if (eptr[i].catch_idx == 0) {
-			info->handler = handler_pc;
-			return (true);
-		}
-		/* Resolve catch class if necessary */
-		if (eptr[i].catch_type == NULL) {
-			eptr[i].catch_type = getClass(eptr[i].catch_idx, ptr->class);
-		}
-		for (cptr = class; cptr != 0; cptr = cptr->superclass) {
-			if (cptr == eptr[i].catch_type) {
-				info->handler = handler_pc;
-				return (true);
-			}
-		}
-	}
-	return (false);
-}
-
-static
-void
-throwNoSuchMethodError(void)
-{
-	throwException(NoSuchMethodError("method name unknown"));
-}
-
 static
 void
 throwAbstractMethodError(void)
