@@ -16,22 +16,40 @@
 #include "config-std.h"
 #include "config-mem.h"
 #include "gtypes.h"
+#include "baseClasses.h"
+#include "support.h"
+#include "stats.h"
 #include "locks.h"
-#include "gc-mem.h"
+#include "jthread.h"
 #include "gc.h"
+#include "gc-mem.h"
 #include "jni.h"
+#ifdef HAVE_UNISTD_H
+#include <unistd.h>
+#endif
 
+extern iLock* gc_lock;
+
+#if defined(KAFFE_STATS)
+static counter gcpages;
+#endif
 static gc_block* gc_small_block(size_t);
 static gc_block* gc_large_block(size_t);
 
 static gc_block* gc_primitive_alloc(size_t);
-static void gc_primitive_free(gc_block*);
+void gc_primitive_free(gc_block*);
 static void* gc_system_alloc(size_t);
 
-static struct {
-	gc_block*	list;
-	uint16		sz;
-} freelist[NR_FREELISTS+1] 
+size_t gc_heap_base;
+size_t gc_block_base;
+uintp gc_heap_range;
+
+typedef struct {
+	gc_block* list;
+	uint16	  sz;
+} gc_freelist;
+
+static gc_freelist freelist[NR_FREELISTS+1]
 #ifdef PREDEFINED_NUMBER_OF_TILES
 	= {
 #define	S(sz)	{ 0, sz }
@@ -54,7 +72,7 @@ static struct {
 	S(1000),
 	S(2016),
 	S(4040),
-	{ (gc_block *)0xffffffff, 0 }
+	{ (gc_block *)((char*)0 - 1), 0 }
 }
 #endif /* PREDEFINED_NUMBER_OF_TILES */
 ;
@@ -66,19 +84,23 @@ static int max_freelist;
 
 static gc_block* gc_prim_freelist;
 static size_t max_small_object_size;
-static iLock gc_lock;
 
-size_t gc_heap_total;
-size_t gc_heap_allocation_size;
-size_t gc_heap_limit;
+size_t gc_heap_total;		/* current size of the heap */
+size_t gc_heap_allocation_size;	/* amount of memory by which to grow heap */
+size_t gc_heap_initial_size;	/* amount of memory to initially allocate */
+size_t gc_heap_limit;		/* maximum size to which heap should grow */
+
+#ifndef gc_pgsize
 size_t gc_pgsize;
-gc_block* gc_objecthash[GC_OBJECT_HASHSIZE];
+int gc_pgbits;
+#endif
+
+#ifdef DEBUG
+int gc_system_alloc_cnt;
+#endif
 
 extern struct Hjava_lang_Thread* garbageman;
 
-static void* pagealloc(size_t);
-extern void throwOutOfMemory(void);
-
 #ifdef DEBUG
 /*
  * analyze the slack incurred by small objects
@@ -86,13 +108,38 @@ extern void throwOutOfMemory(void);
 static int totalslack;
 static int totalsmallobjs;
 
-static
-void printslack()
+static void 
+printslack(void)
 {
-	printf("allocated %d small objects, total slack %d, slack/per "
+	fprintf(stderr, 
+		"allocated %d small objects, total slack %d, slack/per "
 		"object %8.2f\n", 
 		totalsmallobjs, totalslack, totalslack/(double)totalsmallobjs);
 }
+
+/*
+ * check whether the heap is still in a consistent state
+ */
+void
+gc_heap_check(void)
+{
+	int i; 
+
+	for (i = 0; i < NR_FREELISTS; i++) {
+		gc_block* blk = freelist[i].list;
+		if (blk == 0 || blk == (gc_block*)((char*)0 - 1)) {
+			continue;
+		} else {
+			gc_freeobj* mem = blk->free;
+
+			while (mem) {
+				ASSERT_ONBLOCK(mem, blk);
+				mem = mem->next;
+			}
+		}
+	}
+}
+
 #endif /* DEBUG */
 
 /*
@@ -102,21 +149,41 @@ static
 void
 gc_heap_initialise(void)
 {
+#ifndef gc_pgsize
+	gc_pgsize = getpagesize();
+	for (gc_pgbits = 0;
+	     (1 << gc_pgbits) != gc_pgsize && gc_pgbits < 64;
+	     gc_pgbits++)
+		;
+	assert(gc_pgbits < 64);
+#endif
+
+	gc_heap_allocation_size = Kaffe_JavaVMArgs[0].allocHeapSize;
+	gc_heap_initial_size = Kaffe_JavaVMArgs[0].minHeapSize;
+	gc_heap_limit = Kaffe_JavaVMArgs[0].maxHeapSize;
+
+	/*
+	 * Perform some sanity checks.
+	 */
+	if (gc_heap_initial_size > gc_heap_limit) {
+		fprintf(stderr, 
+		    "Initial heap size (%dK) > Maximum heap size (%dK)\n",
+		    (int) (gc_heap_initial_size/1024), (int)(gc_heap_limit/1024));
+		EXIT(-1);
+	}
+
+#ifndef PREDEFINED_NUMBER_OF_TILES
+    {
 	int i;
 	int l;
 	int b;
 	int t;
 
-	gc_pgsize = getpagesize();
-	gc_heap_allocation_size = Kaffe_JavaVMArgs[0].allocHeapSize;
-	gc_heap_limit = Kaffe_JavaVMArgs[0].maxHeapSize;
-
-#ifndef PREDEFINED_NUMBER_OF_TILES
 	/* old scheme, where number of tiles was approximated by a series
 	 * of powers of two
 	 */
 #define	OBJSIZE(NR) \
-	((gc_pgsize-sizeof(gc_block)-ROUNDUPALIGN(1)-(NR*(2+sizeof(void*))))/NR)
+	((gc_pgsize-GCBLOCK_OVH-ROUNDUPALIGN(1)-(NR*(2+sizeof(void*))))/NR)
 
 	/* For a given number of tiles in a block, work out the size of
 	 * the allocatable units which'll fit in them and build a translation
@@ -157,6 +224,7 @@ gc_heap_initialise(void)
 		sztable[l].list = i;
 	}
 	max_freelist = i;
+    }
 #else
 	/* PREDEFINED_NUMBER_OF_TILES */
 	{
@@ -176,14 +244,6 @@ gc_heap_initialise(void)
 	}
 #endif
 
-DBG(GCSTAT,
-	for (i = 0; i < max_small_object_size; i++)
-		printf("size %d list %d, list.size %d\n", i, sztable[i].list,
-			freelist[sztable[i].list].sz);
-	printf("max smobjsize %d, max freelist %d\n", 
-		max_small_object_size, max_freelist);
-    )
-
 DBG(SLACKANAL,
 	atexit(printslack);
     )
@@ -192,6 +252,12 @@ DBG(SLACKANAL,
 
 	/* Round 'gc_heap_allocation_size' up to pagesize */
 	gc_heap_allocation_size = ROUNDUPPAGESIZE(gc_heap_allocation_size);
+
+	/* Round 'gc_heap_initial_size' up to pagesize */
+	gc_heap_initial_size = ROUNDUPPAGESIZE(gc_heap_initial_size);
+
+	/* allocate heap of initial size from system */
+	gc_system_alloc(gc_heap_initial_size);
 }
 
 /*
@@ -202,21 +268,20 @@ void*
 gc_heap_malloc(size_t sz)
 {
 	static int gc_heap_init = 0;
-	static int gc_current_alloc = 0;
 	size_t lnr;
 	gc_freeobj* mem;
 	gc_block** mptr;
 	gc_block* blk;
 	size_t nsz;
 	int times;
+	int iLockRoot;
 
 	/* Initialise GC heap first time in - we must assume single threaded
 	 * operation here so we can do the lock initialising.
 	 */
 	if (gc_heap_init == 0) {
-		gc_heap_init = 1;
 		gc_heap_initialise();
-		initStaticLock(&gc_lock);
+		gc_heap_init = 1;
 	}
 
 	lockStaticMutex(&gc_lock);
@@ -233,6 +298,10 @@ DBG(SLACKANAL,
 	rerun:;
 	times++;
 
+DBG(GCDIAG, 
+	gc_heap_check();
+    )
+
 	if (GC_SMALL_OBJECT(sz)) {
 
 		/* Translate size to object free list */
@@ -244,7 +313,7 @@ DBG(SLACKANAL,
 		if (*mptr != 0) {
 			blk = *mptr;
 			assert(blk->free != 0);
-DBG(GCALLOC,		dprintf("gc_heap_malloc: freelist %d at %p\n", sz, *mptr);)
+DBG(GCALLOC,		dprintf("gc_heap_malloc: freelist %d at %p free %p\n", sz, *mptr, blk->free);)
 		}
 		else {
 			blk = gc_small_block(nsz);
@@ -252,21 +321,20 @@ DBG(GCALLOC,		dprintf("gc_heap_malloc: freelist %d at %p\n", sz, *mptr);)
 				nsz = gc_pgsize;
 				goto nospace;
 			}
-			blk->nfree = *mptr;
+			blk->next = *mptr;
 			*mptr = blk;
 
-DBG(GCALLOC,		dprintf("gc_heap_malloc: small block %d at %p\n", sz, *mptr);)
+DBG(GCALLOC,		dprintf("gc_heap_malloc: small block %d at %p free %p\n", sz, *mptr, blk->free);)
 		}
 
 		/* Unlink free one and return it */
 		mem = blk->free;
-#if defined(GC_DEBUG)
-		assert(blk->magic == GC_MAGIC);
-		assert((uintp)mem >= (uintp)blk &&
-			(uintp)mem < (uintp)blk + gc_pgsize);
-		assert(mem->next == 0 || ((uintp)mem->next >= (uintp)blk &&
-			(uintp)mem->next < (uintp)blk + gc_pgsize));
-#endif
+
+		DBG(GCDIAG,
+		    assert(blk->magic == GC_MAGIC);
+		    ASSERT_ONBLOCK(mem, blk);
+		    if (mem->next) ASSERT_ONBLOCK(mem->next, blk));
+
 		blk->free = mem->next;
 
 		GC_SET_STATE(blk, GCMEM2IDX(blk, mem), GC_STATE_NORMAL);
@@ -277,14 +345,14 @@ DBG(GCALLOC,		dprintf("gc_heap_malloc: small block %d at %p\n", sz, *mptr);)
 		assert(blk->avail > 0);
 		blk->avail--;
 		if (blk->avail == 0) {
-			*mptr = blk->nfree;
+			*mptr = blk->next;
 		}
 	}
 	else {
 		nsz = sz;
 		blk = gc_large_block(nsz);
 		if (blk == 0) {
-			nsz = nsz + sizeof(gc_block) + sizeof(gcFuncs*) + ROUNDUPALIGN(1);
+			nsz = nsz + GCBLOCK_OVH + sizeof(gcFuncs*) + ROUNDUPALIGN(1);
 			nsz = ROUNDUPPAGESIZE(nsz);
 			goto nospace;
 		}
@@ -300,8 +368,6 @@ DBG(GCALLOC,	dprintf("gc_heap_malloc: large block %d at %p\n", sz, mem);	)
 
 	assert(GC_OBJECT_SIZE(mem) >= sz);
 
-	gc_current_alloc += nsz;
-
 	unlockStaticMutex(&gc_lock);
 
 	return (mem);
@@ -313,17 +379,16 @@ DBG(GCALLOC,	dprintf("gc_heap_malloc: large block %d at %p\n", sz, mem);	)
 	 * memory from somewhere.
 	 */
 
-DBG(GCSTAT,
-	dprintf("Demanding %d ...\n", sz);				)
-
 	switch (times) {
 	case 1:
-		/* Try invoking GC, but only if we've got some heap and
-		 * the GC is available and it's worth doing.
-		 */
-		if (gc_heap_total > 0 && garbageman != 0 && gc_current_alloc > gc_heap_allocation_size/2) {
+		/* Try invoking GC if it is available */
+		if (garbageman != 0) {
+			/* The other caller of invokeGC,  Runtime.gc() can't 
+			 * give up this lock on its own, since it does not 
+			 * hold this lock.
+			 */
 			unlockStaticMutex(&gc_lock);
-			invokeGC();
+			adviseGC();
 			lockStaticMutex(&gc_lock);
 		}
 		break;
@@ -333,23 +398,7 @@ DBG(GCSTAT,
 		if (nsz < gc_heap_allocation_size) {
 			nsz = gc_heap_allocation_size;
 		}
-		blk = gc_system_alloc(nsz);
-		if (blk != 0) {
-			/* Place block into the freelist for subsequent use */
-#if defined(GC_DEBUG)
-			blk->magic = GC_MAGIC;
-#endif
-			blk->size = nsz;
-
-			/* Attach block to object hash */
-			lnr = GC_OBJECT_HASHIDX(blk);
-			blk->next = gc_objecthash[lnr];
-			gc_objecthash[lnr] = blk;
-
-			/* Free block into the system */
-			gc_primitive_free(blk);
-		}
-DBG(GCSTAT,	objectStatsPrint();	/* XXX */		)
+		gc_system_alloc(nsz);
 		break;
 
 	default:
@@ -369,7 +418,6 @@ DBG(GCSTAT,	objectStatsPrint();	/* XXX */		)
 	}
 
 	/* Try again */
-	gc_current_alloc = 0;
 	goto rerun;
 }
 
@@ -385,14 +433,15 @@ gc_heap_free(void* mem)
 	int msz;
 	int idx;
 
-	lockStaticMutex(&gc_lock);
 
 	info = GCMEM2BLOCK(mem);
 	idx = GCMEM2IDX(info, mem);
-#if defined(GC_DEBUG)
-	assert(info->magic == GC_MAGIC);
-	assert(GC_GET_COLOUR(info, idx) != GC_COLOUR_FREE);
-#endif
+
+	DBG(GCDIAG,
+	    gc_heap_check();
+	    assert(info->magic == GC_MAGIC);
+	    assert(GC_GET_COLOUR(info, idx) != GC_COLOUR_FREE));
+
 	GC_SET_COLOUR(info, idx, GC_COLOUR_FREE);
 
 DBG(GCFREE,
@@ -404,24 +453,20 @@ DBG(GCFREE,
 		 * it to freelist.
 		 */
 		if (info->avail == 0) {
-			info->nfree = freelist[lnr].list;
+			info->next = freelist[lnr].list;
 			freelist[lnr].list = info;
 		}
 		info->avail++;
-#if defined(GC_DEBUG)
-		/* write pattern in memory to see when live objects were
-		 * freed - Note that (f4f4f4f4 == -185273100)
-		 */
-		memset(mem, 0xf4, info->size);
-#endif
+		DBG(GCDIAG,
+		    /* write pattern in memory to see when live objects were
+		     * freed - Note that (f4f4f4f4 == -185273100)
+		     */
+		    memset(mem, 0xf4, info->size));
 		obj = GCMEM2FREE(mem);
 		obj->next = info->free;
 		info->free = obj;
 
-#if defined(GC_DEBUG)
-		assert((uintp)obj >= (uintp)info &&
-			(uintp)obj < (uintp)info + gc_pgsize);
-#endif
+		ASSERT_ONBLOCK(obj, info);
 
 		/* If we free all sub-blocks, free the block */
 		assert(info->avail <= info->nr);
@@ -429,25 +474,24 @@ DBG(GCFREE,
 			gc_block** finfo = &freelist[lnr].list;
 			for (;;) {
 				if (*finfo == info) {
-					(*finfo) = info->nfree;
+					(*finfo) = info->next;
 					info->size = gc_pgsize;
 					gc_primitive_free(info);
 					break;
 				}
-				finfo = &(*finfo)->nfree;
+				finfo = &(*finfo)->next;
 				assert(*finfo != 0);
 			}
 		}
 	}
 	else {
 		/* Calculate true size of block */
-		msz = info->size + sizeof(gc_block) + ROUNDUPALIGN(1);
+		msz = info->size + GCBLOCK_OVH + ROUNDUPALIGN(1);
 		msz = ROUNDUPPAGESIZE(msz);
 		info->size = msz;
 		gc_primitive_free(info);
 	}
 
-	unlockStaticMutex(&gc_lock);
 }
 
 /*
@@ -468,22 +512,19 @@ gc_small_block(size_t sz)
 	}
 
 	/* Calculate number of objects in this block */
-	nr = (gc_pgsize-sizeof(gc_block)-ROUNDUPALIGN(1))/(sz+2);
+	nr = (gc_pgsize-GCBLOCK_OVH-ROUNDUPALIGN(1))/(sz+2);
 
 	/* Setup the meta-data for the block */
-#if defined(GC_DEBUG)
-	info->magic = GC_MAGIC;
-#endif
+	DBG(GCDIAG, info->magic = GC_MAGIC);
+
 	info->size = sz;
 	info->nr = nr;
 	info->avail = nr;
-	info->funcs = (uint8*)(info + 1);
+	info->funcs = (uint8*)GCBLOCK2BASE(info);
 	info->state = (uint8*)(info->funcs + nr);
 	info->data = (uint8*)ROUNDUPALIGN(info->state + nr);
 
-#if defined(GC_DEBUG)
-	memset(info->data, 0, sz * nr);
-#endif
+	DBG(GCDIAG, memset(info->data, 0, sz * nr));
 
 	/* Build the objects into a free list */
 	for (i = nr-1; i >= 0; i--) {
@@ -512,7 +553,7 @@ gc_large_block(size_t sz)
 	size_t msz;
 
 	/* Add in management overhead */
-	msz = sz+sizeof(gc_block)+2+ROUNDUPALIGN(1);
+	msz = sz+GCBLOCK_OVH+2+ROUNDUPALIGN(1);
 	/* Round size up to a number of pages */
 	msz = ROUNDUPPAGESIZE(msz);
 
@@ -522,20 +563,17 @@ gc_large_block(size_t sz)
 	}
 
 	/* Setup the meta-data for the block */
-#if defined(GC_DEBUG)
-	info->magic = GC_MAGIC;
-#endif
+	DBG(GCDIAG, info->magic = GC_MAGIC);
+
 	info->size = sz;
 	info->nr = 1;
 	info->avail = 1;
-	info->funcs = (uint8*)(info + 1);
+	info->funcs = (uint8*)GCBLOCK2BASE(info);
 	info->state = (uint8*)(info->funcs + 1);
 	info->data = (uint8*)ROUNDUPALIGN(info->state + 1);
 	info->free = 0;
 
-#if defined(GC_DEBUG)
-	memset(info->data, 0, sz);
-#endif
+	DBG(GCDIAG, memset(info->data, 0, sz));
 
 	GCBLOCK2FREE(info, 0)->next = 0;
 	GC_SET_COLOUR(info, 0, GC_COLOUR_FREE);
@@ -544,6 +582,43 @@ gc_large_block(size_t sz)
 	return (info);
 }
 
+/*
+ * Primitive block management:  Allocating and freeing whole pages.
+ *
+ * Unused pages may be marked unreadable.  This is only done when
+ * compiled with DEBUG.
+ */ 
+
+#ifndef PROT_NONE
+#define PROT_NONE 0
+#endif
+
+#if !defined(HAVE_MPROTECT) || !defined(DEBUG)
+#define mprotect(A,L,P)
+#define ALL_PROT
+#define NO_PROT
+#else
+/* In a sense, this is backwards. */
+#define ALL_PROT PROT_READ|PROT_WRITE|PROT_EXEC
+#define NO_PROT  PROT_NONE
+#endif
+
+/* Mark this block as in-use */
+static inline void 
+gc_block_add(gc_block *b)
+{
+	b->inuse = 1;
+	mprotect(GCBLOCK2BASE(b), b->size, ALL_PROT);
+}
+
+/* Mark this block as free */
+static inline void 
+gc_block_rm(gc_block *b)
+{
+	b->inuse = 0;
+	mprotect(GCBLOCK2BASE(b), b->size, NO_PROT);
+}
+
 /*
  * Allocate a block of memory from the free list or, failing that, the
  * system pool.
@@ -554,8 +629,6 @@ gc_primitive_alloc(size_t sz)
 {
 	gc_block* ptr;
 	gc_block** pptr;
-	int hidx;
-	size_t bsz;
 
 	assert(sz % gc_pgsize == 0);
 
@@ -572,17 +645,15 @@ gc_primitive_alloc(size_t sz)
 				ptr->size = sz;
 				nptr = GCBLOCKEND(ptr);
 				nptr->size = left;
-#if defined(GC_DEBUG)
-				nptr->magic = GC_MAGIC;
-#endif
+
+				DBG(GCDIAG, nptr->magic = GC_MAGIC);
+
 				nptr->next = ptr->next;
 				ptr->next = nptr;
 			}
 			*pptr = ptr->next;
 DBG(GCPRIM,		dprintf("gc_primitive_alloc: %d bytes from freelist @ %p\n", ptr->size, ptr); )
-			hidx = GC_OBJECT_HASHIDX(ptr);
-			ptr->next = gc_objecthash[hidx];
-			gc_objecthash[hidx] = ptr;
+			gc_block_add(ptr);
 			return (ptr);
 		}
 	}
@@ -594,31 +665,16 @@ DBG(GCPRIM,		dprintf("gc_primitive_alloc: %d bytes from freelist @ %p\n", ptr->s
 /*
  * Return a block of memory to the free list.
  */
-static
 void
 gc_primitive_free(gc_block* mem)
 {
 	gc_block* lptr;
 	gc_block* nptr;
-	int hidx;
 
 	assert(mem->size % gc_pgsize == 0);
 
 	/* Remove from object hash */
-	hidx = GC_OBJECT_HASHIDX(mem);
-	if (gc_objecthash[hidx] == mem) {
-		gc_objecthash[hidx] = mem->next;
-	}
-	else {
-		for (lptr = gc_objecthash[hidx]; lptr->next != 0; lptr = lptr->next) {
-			if (lptr->next == mem) {
-				lptr->next = mem->next;
-				goto found;
-			}
-		}
-		assert("Failed to find freeing block in object hash" == 0);
-		found:;
-	}
+	gc_block_rm(mem);
 	mem->next = 0;
 
 	if(mem < gc_prim_freelist || gc_prim_freelist == 0) {
@@ -694,39 +750,39 @@ DBG(GCPRIM,	dprintf("gc_primitive_free: Append (%d,%p) onto last in list\n", mem
 	}
 }
 
-static
-void*
-gc_system_alloc(size_t sz)
+/*
+ * Try to reserve some memory for OOM exception handling.  Gc once at
+ * the beginning.  We start out looking for an arbitrary number of
+ * pages (4), and cut our expectations in half until we are able to
+ * meet them.
+ */
+gc_block *
+gc_primitive_reserve(void)
 {
-	void* mem;
-
-	assert(sz % gc_pgsize == 0);
-
-	/* If we will pass the heap boundary, return 0 to indicate that
-	 * we're run out.
-	 */
-	if (gc_heap_total <= gc_heap_limit && gc_heap_total + sz > gc_heap_limit) {
-		gc_heap_total += sz;
-		return (0);
+	gc_block *r = 0;
+	int size = 4 * gc_pgsize;
+	
+	while (size >= gc_pgsize && !(r = gc_primitive_alloc(size))) {
+		if (size == gc_pgsize) {
+			break;
+		}
+		size /= 2;
 	}
-	gc_heap_total += sz;
-
-	mem = pagealloc(sz);
-
-DBG(GCSYSALLOC,
-	dprintf("gc_system_alloc: %d byte at %p\n", sz, mem);		)
-
-	return (mem);
+	return r;
 }
 
-/* --------------------------------------------------------------------- */
-
-void*
+/*
+ * System memory management:  Obtaining additional memory from the
+ * OS.  This looks more complicated than it is, since it does not require
+ * sbrk.
+ */
+/* Get some page-aligned memory from the system. */
+static uintp
 pagealloc(size_t size)
 {
 	void* ptr;
 
-#define	CHECK_OUT_OF_MEMORY(P)	if ((P) == 0) throwOutOfMemory();
+#define	CHECK_OUT_OF_MEMORY(P)	if ((P) == 0) return 0;
 
 #if defined(HAVE_SBRK)
 
@@ -744,6 +800,9 @@ pagealloc(size_t size)
 			break;
 		}
 		missed = gc_pgsize - ((uintp)ptr % gc_pgsize);
+		DBG(GCSYSALLOC,
+		    dprintf("unaligned sbrk %p, missed %d bytes\n",
+			    ptr, missed));
 		sbrk(-size + missed);
 	}
 	CHECK_OUT_OF_MEMORY(ptr);
@@ -764,14 +823,180 @@ pagealloc(size_t size)
 	 * Allocate memory using malloc and align by hand.
 	 */
 	size += gc_pgsize;
-#ifdef malloc
-#undef malloc
-#endif
+
         ptr = malloc(size);
 	CHECK_OUT_OF_MEMORY(ptr);
 	ptr = (void*)((((uintp)ptr) + gc_pgsize - 1) & -gc_pgsize);
 
 #endif
+	addToCounter(&gcpages, "gcmem-system pages", 1, size);
+	return ((uintp) ptr);
+}
 
-	return (ptr);
+/* Free memory allocated with pagealloc */
+static void pagefree(uintp base, size_t size)
+{
+#ifdef HAVE_SBRK
+	sbrk(-size);
+#else
+	/* it must have been allocated with memalign, valloc or malloc */
+	free((void *)base);
+#endif
+}
+
+/*
+ * Allocate size bytes of heap memory, and return the corresponding
+ * gc_block *.
+ */
+static void *
+gc_block_alloc(size_t size)
+{
+	int size_pg = (size>>gc_pgbits);
+	static int n_live = 0;	/* number of pages in java heap */
+	static int nblocks;	/* number of gc_blocks in array */
+	uintp heap_addr;
+	static uintp last_addr;
+
+	if (!gc_block_base) {
+		nblocks = (gc_heap_limit>>gc_pgbits);
+
+		nblocks += nblocks/4;
+		gc_block_base = (uintp) malloc(nblocks * sizeof(gc_block));
+		if (!gc_block_base) return 0;
+		memset((void *)gc_block_base, 0, nblocks * sizeof(gc_block));
+	}
+
+	DBG(GCSYSALLOC, dprintf("pagealloc(%d)", size));
+	heap_addr = pagealloc(size);
+	DBG(GCSYSALLOC, dprintf(" => %p\n", heap_addr));
+
+	if (!heap_addr) return 0;
+	
+	if (!gc_heap_base) {
+		gc_heap_base = heap_addr;
+	}
+
+	if (GCMEM2BLOCK(heap_addr + size)
+	    > ((gc_block *)gc_block_base) + nblocks
+	    || heap_addr < gc_heap_base) {
+		uintp old_blocks = gc_block_base;
+		int onb = nblocks;
+		int min_nb;	/* minimum size of array to hold heap_addr */
+#if defined(KAFFE_STATS)
+		static timespent growtime;
+#endif
+
+		startTiming(&growtime, "gctime-blockrealloc");
+		/* Pick a new size for the gc_block array.  Remember,
+		   malloc does not simply grow a memory segment.
+
+		   We can extrapolate how many gc_blocks we need for
+		   the entire heap based on how many heap pages
+		   currently fit in the gc_block array.  But, we must
+		   also make sure to allocate enough blocks to cover
+		   the current allocation */
+		nblocks = (nblocks * (gc_heap_limit >> gc_pgbits))
+			/ n_live;
+		if (heap_addr < gc_heap_base) 
+			min_nb = nblocks
+			  + ((gc_heap_base - heap_addr) >> gc_pgbits);
+		else
+			min_nb = ((heap_addr + size) - gc_heap_base) >>
+			  gc_pgbits;
+		nblocks = MAX(nblocks, min_nb);
+		DBG(GCSYSALLOC,
+		    dprintf("growing block array from %d to %d elements\n",
+			    onb, nblocks));
+
+		jthread_spinon(0);
+		gc_block_base = (uintp) realloc((void *) old_blocks,
+						nblocks * sizeof(gc_block));
+		if (!gc_block_base) {
+			/* roll back this call */
+			pagefree(heap_addr, size);
+			gc_block_base = old_blocks;
+			nblocks = onb;
+			jthread_spinoff(0);
+			return 0;
+		}
+
+		/* If the array's address has changed, we have to fix
+		   up the pointers in the gc_blocks, as well as all
+		   external pointers to the gc_blocks.  We can only
+		   fix gc_prim_freelist and the size-freelist array.
+		   There should be no gc_block *'s on any stack
+		   now. */ 
+		if (gc_block_base != old_blocks) {
+			extern gc_block *gc_prim_freelist;
+			int i;
+			gc_block *b = (void *) gc_block_base;
+			uintp delta = gc_block_base - old_blocks;
+#define R(X) if (X) ((uintp) (X)) += delta
+
+			DBG(GCSYSALLOC,
+			    dprintf("relocating gc_block array\n"));
+			for (i = 0; i < onb; i++) R(b[i].next);
+			memset(b + onb, 0,
+			       (nblocks - onb) * sizeof(gc_block));
+
+			R(gc_prim_freelist);
+
+			for (i = 0; freelist[i].list != (void *) -1; i++) 
+				R(freelist[i].list);
+#undef R
+		}
+		jthread_spinoff(0);
+		stopTiming(&growtime);
+	}
+	n_live += size_pg;
+	last_addr = MAX(last_addr, heap_addr + size);
+	gc_heap_range = last_addr - gc_heap_base;
+	DBG(GCSYSALLOC, dprintf("%d unused bytes in heap addr range\n",
+				gc_heap_range - gc_heap_total));
+	mprotect((void *) heap_addr, size, NO_PROT);
+	return GCMEM2BLOCK(heap_addr);
+}
+
+static
+void*
+gc_system_alloc(size_t sz)
+{
+	gc_block* blk;
+
+	assert(sz % gc_pgsize == 0);
+
+	if (gc_heap_total == gc_heap_limit) {
+		return (0);
+	} else 	if (gc_heap_total + sz > gc_heap_limit) {
+		/* take as much memory as we can */
+		sz = gc_heap_limit - gc_heap_total;
+		assert(sz % gc_pgsize == 0);
+		DBG(GCSYSALLOC, dprintf("allocating up to limit\n"));
+	}
+#ifdef DEBUG
+	gc_system_alloc_cnt++;
+#endif
+
+	blk = gc_block_alloc(sz);
+	
+DBG(GCSYSALLOC,
+	dprintf("gc_system_alloc: %d byte at %p\n", sz, blk);		)
+
+	if (blk == 0) {
+		return (0);
+	}
+	gc_heap_total += sz;
+	assert(gc_heap_total <= gc_heap_limit);
+
+	/* Place block into the freelist for subsequent use */
+	DBG(GCDIAG, blk->magic = GC_MAGIC);
+	blk->size = sz;
+
+	/* Attach block to object hash */
+	gc_block_add(blk);
+
+	/* Free block into the system */
+	gc_primitive_free(blk);
+
+	return (blk);
 }
