@@ -23,10 +23,10 @@
 #include "access.h"
 #include "object.h"
 #include "constants.h"
+#include "errors.h"
 #include "classMethod.h"
 #include "baseClasses.h"
 #include "lookup.h"
-#include "errors.h"
 #include "exception.h"
 #include "locks.h"
 #include "soft.h"
@@ -34,6 +34,8 @@
 #include "thread.h"
 #include "baseClasses.h"
 #include "itypes.h"
+#include "machine.h"
+#include "fp.h"
 
 /*
  * soft_new
@@ -42,8 +44,11 @@ void*
 soft_new(Hjava_lang_Class* c)
 {
 	Hjava_lang_Object* obj;
+	errorInfo info;
 
-	processClass(c, CSTATE_OK);
+	if (processClass(c, CSTATE_COMPLETE, &info) == false) {
+		throwError(&info);
+	}
 	obj = newObject(c);
 
 DBG(NEWINSTR,	
@@ -113,7 +118,7 @@ soft_multianewarray(Hjava_lang_Class* class, jint dims, slots* args)
 		arraydims = array;
 	}
 	else {
-		arraydims = gc_calloc_fixed(dims+1, sizeof(int));
+		arraydims = checkPtr(KCALLOC(dims+1, sizeof(int)));
 	}
 
 	/* stack grows up, so move to the first dimension */
@@ -127,14 +132,14 @@ soft_multianewarray(Hjava_lang_Class* class, jint dims, slots* args)
 		}
                 arraydims[i] = arg;
         }
-        arraydims[i] = 0;
+        arraydims[i] = -1;
 
         /* Mmm, okay now build the array using the wonders of recursion */
         obj = newMultiArray(class, arraydims);
 
 
 	if (arraydims != array) {
-		gc_free_fixed(arraydims);
+		KFREE(arraydims);
 	}
 
         /* Return the base object */
@@ -157,7 +162,7 @@ soft_multianewarray(Hjava_lang_Class* class, jint dims, ...)
 		arraydims = array;
 	}
 	else {
-		arraydims = gc_calloc_fixed(dims+1, sizeof(int));
+		arraydims = checkPtr(KCALLOC(dims+1, sizeof(int)));
 	}
 
 	/* Extract the dimensions into an array */
@@ -165,18 +170,22 @@ soft_multianewarray(Hjava_lang_Class* class, jint dims, ...)
 	for (i = 0; i < dims; i++) {
 		arg = va_arg(ap, jint);
 		if (arg < 0) {
+			va_end(ap);
+			if (arraydims != array) {
+				KFREE(arraydims);
+			}
                         throwException(NegativeArraySizeException);
 		}
 		arraydims[i] = arg;
 	}
-	arraydims[i] = 0;
+	arraydims[i] = -1;
 	va_end(ap);
 
 	/* Mmm, okay now build the array using the wonders of recursion */
         obj = newMultiArray(class, arraydims);
 
 	if (arraydims != array) {
-		gc_free_fixed(arraydims);
+		KFREE(arraydims);
 	}
 
 	/* Return the base object */
@@ -188,22 +197,77 @@ soft_multianewarray(Hjava_lang_Class* class, jint dims, ...)
  * soft_lookupmethod.
  */
 void*
-soft_lookupmethod(Hjava_lang_Object* obj, Utf8Const* name, Utf8Const* sig)
+soft_lookupmethod(Hjava_lang_Object* obj, Hjava_lang_Class* ifclass, int idx)
 {
 	Hjava_lang_Class* cls;
+	Hjava_lang_Class** c;
 	Method* meth;
+	void*	ncode;
+	int i;
+	int dtableidx;
+	errorInfo info;
 
-	cls = OBJECT_CLASS(obj);
-	meth = findMethod(cls, name, sig);
-	if (meth == 0) {
-		throwException(NoSuchMethodError(name->data));
+	if (obj == 0) {
+		throwException(NullPointerException);
 	}
 
-#if defined(TRANSLATOR)
-	return (METHOD_NATIVECODE(meth));
-#else
-	return (meth);
-#endif
+	cls = OBJECT_CLASS(obj);
+	/* initialize class if necessary */
+        if (cls->state < CSTATE_USABLE) {
+		if (processClass(cls, CSTATE_COMPLETE, &info) == false) {
+			throwError(&info);
+		}
+	}
+
+	c = cls->interfaces;
+	i = 0;
+	while (*c != ifclass) {
+		c++;
+		/* do this check inside the loop to avoid loading 
+		 * total_interface_len from memory if we hit the first one. 
+		 */
+		if (++i == cls->total_interface_len) {
+			goto notfound;
+		}
+	}
+
+	dtableidx = cls->itable2dtable[cls->if2itable[i] + idx];
+	if (dtableidx == -1) {
+		goto notfound;
+	}
+
+	ncode = *(void**)((char*)cls->dtable 
+			+ DTABLE_METHODOFFSET + dtableidx * DTABLE_METHODSIZE);
+	return (ncode);
+
+notfound:
+	/*
+	 * Compilers following the latest version of the JLS emit a 
+	 * INVOKEINTERFACE instruction for methods that aren't defined in 
+	 * an interface, but inherited from Object.
+	 *
+	 * In this case, the JVM must
+	 * a) check that the object really implements the interface given
+	 * b) find and invoke the method on object.
+	 *
+	 * The best way to jit that would be a checkcast <interface_type>
+	 * followed by an INVOKEVIRTUAL.
+	 *
+	 * For now, we simply detect the case where an object method is called
+	 * and find it by hand using its (name, signature). 
+	 */
+	if (ifclass == ObjectClass) {
+		Method* objm = CLASS_METHODS(ifclass) + idx;
+		
+		meth = findMethod(cls, objm->name, METHOD_SIG(objm), &info);
+		if (meth == 0) {
+			throwError(&info);
+		}
+		return (METHOD_NATIVECODE(meth));
+	}
+	meth = CLASS_METHODS(ifclass) + idx;
+	soft_nosuchmethod(cls, meth->name, METHOD_SIG(meth));
+	return (0);
 }
 
 #if 0
@@ -244,7 +308,6 @@ jint
 instanceof_class(Hjava_lang_Class* c, Hjava_lang_Class* oc)
 {
 	Hjava_lang_Class* tc;
-	int i;
 
 	/* Check for superclass matches */
 	for (tc = oc->superclass; tc != 0; tc = tc->superclass) {
@@ -259,7 +322,6 @@ inline
 jint
 instanceof_interface(Hjava_lang_Class* c, Hjava_lang_Class* oc)
 {
-	Hjava_lang_Class* tc;
 	int i;
 
 	/* Check 'total' interface list */
@@ -350,7 +412,20 @@ void*
 soft_checkcast(Hjava_lang_Class* c, Hjava_lang_Object* o)
 {
 	if (o != 0 && !instanceof(c, OBJECT_CLASS(o))) {
-		throwException(ClassCastException);
+		/* 
+		 * Let's be a bit more informative as to why the class 
+		 * cast exception happened.
+		 */
+		Hjava_lang_Throwable* ccexc;
+		const char *fromtype = CLASS_CNAME(OBJECT_CLASS(o));
+		const char *totype = CLASS_CNAME(c);
+		char *format = "can't cast `%s' to `%s'";
+		char *buf = checkPtr(KMALLOC(strlen(fromtype) 
+			+ strlen(totype) + strlen(format)));
+		sprintf(buf, format, fromtype, totype);
+		ccexc = ClassCastException(buf);
+		KFREE(buf);
+		throwException(ccexc);
 	}
 	return (o);
 }
@@ -366,7 +441,14 @@ soft_athrow(Hjava_lang_Object* o)
 		soft_nullpointer();
 	}
 	else {
-		throwExternalException(o);
+		/* NB: This will keep the stacktrace that was built
+		 * when the exception was constructed.
+		 * If you wanted a new stacktrace corresponding to the
+		 * site where the exception is thrown, you'd use 
+		 * `throwException' instead.  However, this would slow down 
+		 * exceptions because two stacktrace have to be constructed.
+		 */
+		throwExternalException((Hjava_lang_Throwable*)o);
 	}
 }
 
@@ -388,17 +470,60 @@ soft_nullpointer(void)
 	throwException(NullPointerException);
 }
 
+/*
+ * soft_divzero.
+ */
+void
+soft_divzero(void)
+{
+	throwException(ArithmeticException);
+}
+
+/*
+ * soft_stackoverflow.
+ */
+void
+soft_stackoverflow(void)
+{
+	throwException(unhand(getCurrentThread())->stackOverflowError);
+}
+
+/*
+ * soft_nosuchclass.
+ */
+void            
+soft_nosuchclass(Utf8Const* c)
+{
+	char buf[256];
+
+	sprintf(buf, "%.80s", c->data);
+	throwException(NoClassDefFoundError(buf));
+}
+
 /*
  * soft_nosuchmethod.
  */
 void            
 soft_nosuchmethod(Hjava_lang_Class* c, Utf8Const* n, Utf8Const* s)
 {
-	char buf[100];
-	sprintf(buf, "%s.%s%s", CLASS_CNAME(c), n->data, s->data);
+	char buf[256];
+
+	sprintf(buf, "%.80s.%.80s%.80s", CLASS_CNAME(c), n->data, s->data);
 	throwException(NoSuchMethodError(buf));
 }
 
+/*
+ * soft_nosuchfield.
+ */
+void
+soft_nosuchfield(Utf8Const* c, Utf8Const* n)
+{
+	char buf[256];
+
+	sprintf(buf, "%.100s.%.100s", c->data, n->data);
+	throwException(NoSuchFieldError(buf));
+}
+
 /*
  * soft_initialise_class.
  */
@@ -406,8 +531,11 @@ void
 soft_initialise_class(Hjava_lang_Class* c)
 {
 	/* We check this outside the processClass to save a subroutine call */
-	if (c->state != CSTATE_OK) {
-		processClass(c, CSTATE_OK);
+	if (c->state != CSTATE_COMPLETE) {
+		errorInfo info;
+		if (processClass(c, CSTATE_COMPLETE, &info) == false) {
+			throwError(&info);
+		}
 	}
 }
 
@@ -419,15 +547,20 @@ nativecode*
 soft_fixup_trampoline(FIXUP_TRAMPOLINE_DECL)
 {
 	Method* meth;
-	nativecode* ocode;
+	errorInfo info;
 	FIXUP_TRAMPOLINE_INIT;
 
 	/* If this class needs initializing, do it now.  */
-	processClass(meth->class, CSTATE_OK);
+	if (meth->class->state < CSTATE_USABLE &&
+		processClass(meth->class, CSTATE_COMPLETE, &info) == false) {
+		throwError(&info);
+	}
 
 	/* Generate code on demand.  */
 	if (!METHOD_TRANSLATED(meth)) {
-		translate(meth);
+		if (translate(meth, &info) == false) {
+			throwError(&info);
+		}
 	}
 
 	/* Update dispatch table */
@@ -438,7 +571,7 @@ soft_fixup_trampoline(FIXUP_TRAMPOLINE_DECL)
 #if 0
 	if (METHOD_PRE_COMPILED(meth)) {
 		nativecode* ncode = METHOD_TRUE_NCODE(meth);
-		ocode = METHOD_NATIVECODE(meth);
+		nativecode* ocode = METHOD_NATIVECODE(meth);
 		METHOD_NATIVECODE(meth) = ncode;
 		/* Update the dtable entries for all classes if this isn't a
 	   	   static method.  */
@@ -449,7 +582,7 @@ soft_fixup_trampoline(FIXUP_TRAMPOLINE_DECL)
 	}
 #endif
 
-TDBG(	fprintf(stderr, "Calling %s:%s%s @ 0x%x\n", meth->class->name->data, meth->name->data, meth->signature->data, METHOD_NATIVECODE(meth));	)
+TDBG(	fprintf(stderr, "Calling %s:%s%s @ 0x%x\n", meth->class->name->data, meth->name->data, METHOD_SIGD(meth), METHOD_NATIVECODE(meth));	)
 
 	return (METHOD_NATIVECODE(meth));
 }
@@ -518,7 +651,13 @@ jint
 soft_fcmpg(jfloat v1, jfloat v2)
 {
         jint ret;
-	if ((!isinf(v1) && isnan(v1)) || (!isinf(v2) && isnan(v2))) {
+	jint v1bits;
+	jint v2bits;
+
+	v1bits = floatToInt(v1);
+	v2bits = floatToInt(v2);
+
+        if (FISNAN(v1bits) || FISNAN(v2bits)) {
 		ret = 1;
 	}
         else if (v1 > v2) {
@@ -540,7 +679,13 @@ jint
 soft_fcmpl(jfloat v1, jfloat v2)
 {
         jint ret;
-	if ((!isinf(v1) && isnan(v1)) || (!isinf(v2) && isnan(v2))) {
+	jint v1bits;
+	jint v2bits;
+
+	v1bits = floatToInt(v1);
+	v2bits = floatToInt(v2);
+
+        if (FISNAN(v1bits) || FISNAN(v2bits)) {
 		ret = -1;
 	}
         else if (v1 > v2) {
@@ -573,48 +718,65 @@ soft_lrem(jlong v1, jlong v2)
 	return (v1 % v2);
 }
 
+jfloat
+soft_fadd(jfloat v1, jfloat v2)
+{
+	return floatAdd(v1, v2);
+}
+
+jdouble
+soft_faddl(jdouble v1, jdouble v2)
+{
+	return doubleAdd(v1, v2);
+}
+
+jfloat
+soft_fsub(jfloat v1, jfloat v2)
+{
+	return floatSubtract(v1, v2);
+}
+
+jdouble
+soft_fsubl(jdouble v1, jdouble v2)
+{
+	return doubleSubtract(v1, v2);
+}
+
+
+jfloat
+soft_fmul(jfloat v1, jfloat v2)
+{
+	return floatMultiply(v1, v2);
+}
+
+jdouble
+soft_fmull(jdouble v1, jdouble v2)
+{
+	return doubleMultiply(v1, v2);
+}
+
 jfloat
 soft_fdiv(jfloat v1, jfloat v2)
 {
-	jfloat k = 1e300;
-
-	if (v2 == 0.0) {
-		if (v1 > 0) {
-			return (k*k);
-		}
-		else {
-			return (-k*k);
-		}
-	}
-	return (v1 / v2);
+	return floatDivide(v1, v2);
 }
 
 jdouble
 soft_fdivl(jdouble v1, jdouble v2)
 {
-	jdouble k = 1e300;
-
-	if (v2 == 0.0) {
-		if (v1 > 0) {
-			return (k*k);
-		}
-		else {
-			return (-k*k);
-		}
-	}
-	return (v1 / v2);
+	return doubleDivide(v1, v2);
 }
 
 jfloat
 soft_frem(jfloat v1, jfloat v2)
 {
-	return (remainderf(v1, v2));
+	return (javaRemainderf(v1, v2));
 }
 
 jdouble
 soft_freml(jdouble v1, jdouble v2)
 {
-	return (remainder(v1, v2));
+	return (javaRemainder(v1, v2));
 }
 
 jlong
@@ -706,13 +868,29 @@ soft_cvtld(jlong v)
 jdouble
 soft_cvtfd(jfloat v)
 {
-	return ((jdouble)v);
+	jint vbits;
+
+	vbits = floatToInt(v);
+        if (FISNAN(vbits)) {
+		return (longToDouble(DNANBITS));
+	}
+	else {
+		return ((jdouble)v);
+	}
 }
 
 jfloat
 soft_cvtdf(jdouble v)
 {
-	return ((jfloat)v);
+	jlong vbits;
+
+	vbits = doubleToLong(v);
+        if (DISNAN(vbits)) {
+		return (intToFloat(FNANBITS));
+	}
+	else {
+		return ((jfloat)v);
+	}
 }
 
 /*
@@ -745,21 +923,72 @@ soft_cvtdl(jdouble v)
 jint
 soft_cvtfi(jfloat v)
 {
+        jint vbits;
+
+	vbits = floatToInt(v);
+        if (FISNAN(vbits)) {
+		return (0);
+	}
+
 	if (v < 0.0) {
-		return ((jint)ceil(v));
+		v = ceil(v);
 	}
 	else {
-		return ((jint)floor(v));
+		v = floor(v);
+	}
+	/* If too small return smallest int */
+	if (v < -2147483648.0) {
+		return (-2147483647-1);
+	}
+	/* If too big return biggest int */
+	else if (v > 2147483647) {
+		return (2147483647);
+	}
+	else {
+		return ((jint)v);
 	}
 }
 
 jint
 soft_cvtdi(jdouble v)
 {
+        jlong vbits;
+
+	vbits = doubleToLong(v);
+        if (DISNAN(vbits)) {
+		return (0);
+	}
+
 	if (v < 0.0) {
-		return ((jint)ceil(v));
+		v = ceil(v);
 	}
 	else {
-		return ((jint)floor(v));
+		v = floor(v);
+	}
+	/* If too small return smallest int */
+	if (v < -2147483648.0) {
+		return (-2147483647-1);
+	}
+	/* If too big return biggest int */
+	else if (v > 2147483647) {
+		return (2147483647);
+	}
+	else {
+		return ((jint)v);
 	}
 }
+
+void
+soft_debug1(void* a0, void* a1, void* a2)
+{
+}
+
+void
+soft_debug2(void* a0, void* a1, void* a2)
+{
+}
+
+void
+soft_trace(Method* meth, void* args)
+{
+}
