@@ -9,13 +9,12 @@
  * of this file. 
  */
 
-#define	DBG(s)
-
 #include "config.h"
 #include "config-std.h"
 #include "config-signal.h"
 #include "config-mem.h"
 #include "config-setjmp.h"
+#include "debug.h"
 #include "jtypes.h"
 #include "access.h"
 #include "object.h"
@@ -32,105 +31,21 @@
 #include "soft.h"
 #include "md.h"
 #include "stackTrace.h"
+#include "support.h"
 
-#define	ENDOFSTACK		((Method*)-1)
-typedef struct _stackTraceInfo {
-	uintp	pc;
-	Method*	meth;
-} stackTraceInfo;
 
-Hjava_lang_Object*
-getClassContext(void* bulk)
-{
-	int cnt=0;
-	int i;
-	stackTrace trace;
-	Hjava_lang_Object* clazz;
-	Method* meth;
-
-	STACKTRACEINIT(trace,NULL,bulk);
-	for( ; !STACKTRACEEND(trace) ; STACKTRACESTEP(trace) ) {
-		if ( STACKTRACEMETH(trace) != 0 ) {
-			cnt++;
-		}
-	}
-
-	assert( cnt > 0 );
-
-	clazz = newArray(ClassClass, cnt);
-
-	STACKTRACEINIT(trace,NULL,bulk);
-	for(i=0;!STACKTRACEEND(trace);STACKTRACESTEP(trace)){
-		meth=STACKTRACEMETH(trace);
-		if ( meth !=0 ) {
-			OBJARRAY_DATA(clazz)[i] =
-				(Hjava_lang_Object*)meth->class;
-			fprintf(stderr,"class %s, loader %p\n",
-				CLASS_CNAME(meth->class),meth->class->loader);
-			i++;
-		}
-	}
-
-	assert( cnt == i );
-
-	return clazz;
-}
-
-Hjava_lang_Class*
-getClassWithLoader(int* depth)
-{
-	int cnt;
-	stackTrace trace;
-	Method* meth;
-
-	cnt = 0;
-	STACKTRACEINIT(trace, NULL, depth);
-	for(; !STACKTRACEEND(trace); STACKTRACESTEP(trace)) {
-		meth = STACKTRACEMETH(trace);
-		if (meth != 0) {
-			if (meth->class->loader != NULL) {
-				*depth = cnt;
-				return (meth->class);
-			}
-			cnt++;
-		}
-	}
-	*depth = -1;
-	return (NULL);
-}
-
-jint
-classDepth(char* name)
-{
-	int cnt;
-	stackTrace trace;
-	Method* meth;
-
-	cnt = 0;
-	STACKTRACEINIT(trace,NULL,name);
-	for( ; !STACKTRACEEND(trace); STACKTRACESTEP(trace) ) {
-		meth = STACKTRACEMETH(trace);
-		if (meth != 0) {
-			if (!strcmp(CLASS_CNAME(meth->class), name)) {
-				return (cnt);
-			}
-			cnt++;
-		}
-	}
-	return(-1);
-}
-
-/*
- * Build an array of stackTraceInfo[] for the current stack backtrace.
- */
 Hjava_lang_Object*
 buildStackTrace(struct _exceptionFrame* base)
 {
 	int cnt;
 	struct _stackTrace trace;
 	stackTraceInfo* info;
+	struct _exceptionFrame orig;
 
-	STACKTRACEINIT(trace, base, base);
+	(void) orig;			/* avoid compiler warning in intrp */
+	DBG(STACKTRACE,
+	    dprintf("STACKTRACEINIT(trace, base, base, orig);\n"); );
+	STACKTRACEINIT(trace, base, base, orig);
 	cnt = 0;
 	while(!STACKTRACEEND(trace)) {
 		STACKTRACESTEP(trace);
@@ -139,19 +54,18 @@ buildStackTrace(struct _exceptionFrame* base)
 
 	/* Build an array of stackTraceInfo */
 	info = gc_malloc(sizeof(stackTraceInfo) * (cnt+1), GC_ALLOC_NOWALK);
+	if (!info) return 0;
 
 	cnt = 0;
 
-	STACKTRACEINIT(trace,base,base);
+	DBG(STACKTRACE,
+	    dprintf("STACKTRACEINIT(trace, &orig, base, orig);\n"); );
+	STACKTRACEINIT(trace, &orig, base, orig);
 
 	for(; !STACKTRACEEND(trace); STACKTRACESTEP(trace)) {
 		info[cnt].pc = STACKTRACEPC(trace);
-#if defined(INTERPRETER)
-		info[cnt].meth = STACKTRACEMETH(trace);
-#endif
-#if defined(TRANSLATOR)
-		info[cnt].meth = 0;	/* We do this lazily */
-#endif
+		info[cnt].fp = STACKTRACEFP(trace);
+		info[cnt].meth = STACKTRACEMETHCREATE(trace);
 		cnt++;
 	}
 	info[cnt].pc = 0;
@@ -160,8 +74,28 @@ buildStackTrace(struct _exceptionFrame* base)
 	return ((Hjava_lang_Object*)info);
 }
 
+/*
+ * We make these exported functions because we use them in Class.c
+ */
+#if defined(TRANSLATOR)
+Method*
+stacktraceFindMethod(stackTraceInfo *info)
+{
+	return (findMethodFromPC(info->pc));
+}
+
+#elif defined(INTERPRETER)
+
+Method*
+stacktraceFindMethod(stackTraceInfo *info)
+{
+	return (info->meth);
+}
+#endif
+
 void
-printStackTrace(struct Hjava_lang_Throwable* o, struct Hjava_lang_Object* p)
+printStackTrace(struct Hjava_lang_Throwable* o,
+		struct Hjava_lang_Object* p, int nullOK)
 {
 	int i;
 	stackTraceInfo* info;
@@ -169,11 +103,12 @@ printStackTrace(struct Hjava_lang_Throwable* o, struct Hjava_lang_Object* p)
 	uintp pc;
 	int32 linenr;
 	uintp linepc;
-	char buf[200];
+	char *buf;
 	int len;
 	int j;
 	Hjava_lang_Object* str;
 	jchar* cptr;
+	char* class_dot_name;
 
 	info = (stackTraceInfo*)unhand(o)->backtrace;
 	if (info == 0) {
@@ -181,42 +116,69 @@ printStackTrace(struct Hjava_lang_Throwable* o, struct Hjava_lang_Object* p)
 	}
 	for (i = 0; info[i].meth != ENDOFSTACK; i++) {
 		pc = info[i].pc;
-#if defined(INTERPRETER)
-		meth = info[i].meth; 
-#endif
-#if defined(TRANSLATOR)
-		meth = findMethodFromPC(pc);
-#endif
+		meth = stacktraceFindMethod(&info[i]);
 		if (meth != 0) {
 			linepc = 0;
 			linenr = -1;
 			if (meth->lines != 0) {
 				for (j = 0; j < meth->lines->length; j++) {
-					if (pc >= meth->lines->entry[j].start_pc && linepc < meth->lines->entry[j].start_pc) {
+					if (pc >= meth->lines->entry[j].start_pc && linepc <= meth->lines->entry[j].start_pc) {
 						linenr = meth->lines->entry[j].line_nr;
 						linepc = meth->lines->entry[j].start_pc;
 					}
 				}
 			}
+			/* Even if we are reporting an out of memory and
+			   checkPtr fails, this is ok.  If we can't allocate
+			   a new one, the vm will die in an orderly manner.
+			*/
+			class_dot_name = checkPtr(
+			    KMALLOC(strlen(CLASS_CNAME(meth->class)) + 1));
+			pathname2classname(CLASS_CNAME(meth->class), class_dot_name);
+			buf = checkPtr(KMALLOC(strlen(class_dot_name)
+			    + strlen(meth->name->data)
+			    + strlen(CLASS_SOURCEFILE(meth->class))
+			    + 64));
+
 			if (linenr == -1) {
-				sprintf(buf, "\tat %.80s.%.80s(line unknown, pc %p)",
-					CLASS_CNAME(meth->class),
-					meth->name->data, (void*)pc);
+				if (meth->accflags & ACC_NATIVE) {
+					sprintf(buf, "\tat %s.%s(%s:native)",
+						class_dot_name,
+						meth->name->data, 
+						CLASS_SOURCEFILE(meth->class));
+				}
+				else {
+					sprintf(buf, "\tat %s.%s(%s:line unknown, pc %p)",
+						class_dot_name,
+						meth->name->data, 
+						CLASS_SOURCEFILE(meth->class),
+						(void*)pc);
+				}
 			}
 			else {
-				sprintf(buf, "\tat %.80s.%.80s(%d)",
-					CLASS_CNAME(meth->class),
+				sprintf(buf, "\tat %s.%s(%s:%d)",
+					class_dot_name,
 					meth->name->data,
+					CLASS_SOURCEFILE(meth->class),
 					linenr);
 			}
+			KFREE(class_dot_name);
 			len = strlen(buf);
 			str = newArray(TYPE_CLASS(TYPE_Char), len);
 			cptr = (jchar*)OBJARRAY_DATA(str);
 			for (j = len;  --j >= 0; ) {
 				cptr[j] = (unsigned char)buf[j];
 			}
-			do_execute_java_method(p,"println","([C)V",0,0,str);
+			if (p != 0 || !nullOK) {
+				do_execute_java_method(p, "println",
+					"([C)V", 0, 0, str);
+			} else {
+				fprintf(stderr, "%s\n", buf);
+			}
+			KFREE(buf);
 		}
 	}
-	do_execute_java_method(p, "flush", "()V", 0, 0);
+	if (p != 0 || !nullOK) {
+		do_execute_java_method(p, "flush", "()V", 0, 0);
+	}
 }
