@@ -12,6 +12,7 @@
 #ifndef __support_h
 #define	__support_h
 
+#include "config-std.h"
 #include <stdarg.h>
 
 /* For user defined properties */
@@ -33,6 +34,7 @@ extern nativeFunction* native_funcs;
 extern userProperty* userProperties;
 
 struct _methods;
+struct _errorInfo;
 
 /* 64 should put us on the safe side */
 #define	MAXMARGS		64
@@ -70,8 +72,14 @@ struct _methods;
  * in args[i].  Similarly, calltype[i] contains the signature type of the
  * argument in args[i] ('J', 'F', etc.)
  *
- * Porters note that "callsize[i] == 2 iff callsize[i+1] == 0" -- this 
- * property is exploited by some sysdepCallMethod macros (sparc).
+ * To simplify 32 bit assembly code, we copy the last 32 bits of a 64
+ * bit arg into the next slot.  This allows you to treat args as an
+ * array of 32 bit values.  This simplification also makes a C version
+ * of sysdepCallMethod more viable, and such a function is defined in
+ * mi.c.
+ *
+ * Note that "callsize[i] == 2 iff callsize[i+1] == 0" -- this 
+ * property can also be exploited by a sysdepCallMethod macros. 
  *
  * `function' is a pointer to the method to be invoked.
  *
@@ -93,33 +101,66 @@ typedef struct {
 
 struct Hjava_lang_String;
 struct Hjava_lang_Class;
+struct Hjava_lang_Object;
+struct Hjava_lang_Throwable;
 
 extern void		setProperty(void*, char*, char*);
-extern void		classname2pathname(char*, char*);
-extern struct Hjava_lang_Object* makeJavaCharArray(char*, int);
-extern struct Hjava_lang_String* makeJavaString(char*, int);
-extern char*		javaString2CString(struct Hjava_lang_String*, char*, int);
-extern char*		makeCString(struct Hjava_lang_String*);
-extern struct Hjava_lang_String* makeReplaceJavaStringFromUtf8(unsigned char*, int, int, int);
-extern jvalue		do_execute_java_method(void*, char*, char*, struct _methods*, int, ...);
-extern jvalue		do_execute_java_method_v(void*, char*, char*, struct _methods*, int, va_list);
-extern jvalue		do_execute_java_class_method(char*, char*, char*, ...);
-extern jvalue		do_execute_java_class_method_v(char*, char*, char*, va_list);
+extern char*		getEngine(void);
+extern void		classname2pathname(const char*, char*);
+extern void		pathname2classname(const char*, char*);
+extern jvalue		do_execute_java_method(void*, const char*, const char*, struct _methods*, int, ...);
+extern jvalue		do_execute_java_method_v(void*, const char*, const char*, struct _methods*, int, va_list);
+extern jvalue		do_execute_java_class_method(const char*, const char*, const char*, ...);
+extern jvalue		do_execute_java_class_method_v(const char*, const char*, const char*, va_list);
 
-extern Hjava_lang_Object* execute_java_constructor(char*, struct Hjava_lang_Class*, char*, ...);
-extern Hjava_lang_Object* execute_java_constructor_v(char*, struct Hjava_lang_Class*, char*, va_list);
+extern struct Hjava_lang_Object* execute_java_constructor(const char*, struct Hjava_lang_Class*, const char*, ...);
+extern struct Hjava_lang_Object* execute_java_constructor_v(const char*, struct Hjava_lang_Class*, const char*, va_list);
 extern jlong		currentTime(void);
-extern void		addNativeMethod(char*, void*);
+extern void		addNativeMethod(const char*, void*);
 
-extern void	callMethodA(struct _methods*, void*, void*, jvalue*, jvalue*);
+extern void	callMethodA(struct _methods*, void*, void*, jvalue*, jvalue*, int);
 extern void	callMethodV(struct _methods*, void*, void*, va_list, jvalue*);
-extern struct _methods*	lookupClassMethod(struct Hjava_lang_Class*, char*, char*);
-extern struct _methods*	lookupObjectMethod(struct Hjava_lang_Object*, char*, char*);
+extern struct _methods*	lookupClassMethod(struct Hjava_lang_Class*, const char*, const char*, struct _errorInfo*);
+extern struct _methods*	lookupObjectMethod(struct Hjava_lang_Object*, const char*, const char*, struct _errorInfo*);
 
 struct _strconst;
-extern void SignalError(char *, char *) __NORETURN__;
-extern void unimp(char*) __NORETURN__;
-extern struct Hjava_lang_String* Utf8Const2JavaString(struct _strconst*);
+extern void SignalError(const char *, const char *) __NORETURN__;
+extern void SignalErrorf(const char *, const char *, ...) __NORETURN__;
+extern void unimp(const char*) __NORETURN__;
 extern void kprintf(FILE*, const char*, ...);
+extern int addClasspath(const char*);
+extern int prependClasspath(const char*);
+
+/*
+ * Macros to manipulate bit arrays.  
+ */
+#define BITMAP_BPI	(sizeof(int) * 8)
+
+/* create a new bitmap for b bits */
+#define BITMAP_NEW(b)	\
+    (int *)KCALLOC(((b) + BITMAP_BPI - 1)/BITMAP_BPI, sizeof(int))
+
+/* set nth bit, counting from MSB to the right */
+#define BITMAP_SET(m, n) \
+    (m[(n) / BITMAP_BPI] |= (1 << (BITMAP_BPI - 1 - (n) % BITMAP_BPI)))
+
+/* copy nbits from bitmap src to bitmap dst */
+#define BITMAP_COPY(dst, src, nbits) \
+    memcpy(dst, src, ((nbits + BITMAP_BPI - 1)/BITMAP_BPI) * sizeof(int))
+
+/* dump a bitmap for debugging */
+#define BITMAP_DUMP(m, N) { int n; \
+    for (n = 0; n < N; n++) \
+	if (m[(n) / BITMAP_BPI] &  \
+		(1 << (BITMAP_BPI - 1 - (n) % BITMAP_BPI))) \
+	    dprintf("1"); else dprintf("0"); }
+
+#ifndef MIN
+#define MIN(A,B) ((A) < (B) ? (A) : (B))
+#endif
+
+#ifndef MAX
+#define MAX(A,B) ((A) > (B) ? (A) : (B))
+#endif
 
 #endif
