@@ -4,7 +4,7 @@
  * Copyright (c) 1996, 1997
  *	Transvirtual Technologies, Inc.  All rights reserved.
  *
- * See the file "lib-license.terms" for information on usage and redistribution 
+ * See the file "license.terms" for information on usage and redistribution 
  * of this file. 
  */
 
@@ -15,15 +15,17 @@
 #include "../../../kaffe/kaffevm/classMethod.h"
 #include "../../../kaffe/kaffevm/baseClasses.h"
 #include "../../../kaffe/kaffevm/itypes.h"
-#include "Array.h"
-#include "Boolean.h"
-#include "Byte.h"
-#include "Character.h"
-#include "Short.h"
-#include "Integer.h"
-#include "Long.h"
-#include "Float.h"
-#include "Double.h"
+#include "../../../kaffe/kaffevm/soft.h"
+#include "../../../kaffe/kaffevm/exception.h"
+#include "java_lang_reflect_Array.h"
+#include "java_lang_Boolean.h"
+#include "java_lang_Byte.h"
+#include "java_lang_Character.h"
+#include "java_lang_Short.h"
+#include "java_lang_Integer.h"
+#include "java_lang_Long.h"
+#include "java_lang_Float.h"
+#include "java_lang_Double.h"
 #include <native.h>
 #include "defs.h"
 
@@ -46,69 +48,69 @@ java_lang_reflect_Array_get(struct Hjava_lang_Object* obj, jint elem)
 
 	clazz = CLASS_ELEMENT_TYPE(OBJECT_CLASS(obj));
 	
-	if ( clazz == intClass ) {
+	if ( clazz == _Jv_intClass ) {
 		HArrayOfInt *arr = (HArrayOfInt *)obj;
 		if (elem < 0 || elem >= obj_length(arr))
 			SignalError("java.lang.ArrayIndexOutOfBoundsException", "");
 		
 		return execute_java_constructor(0,javaLangIntegerClass,"(I)V",
-						(jint)(unhand(arr)->body[elem]));
+						(jint)(unhand_array(arr)->body[elem]));
 	}
-	else if ( clazz == shortClass ) {
+	else if ( clazz == _Jv_shortClass ) {
 		HArrayOfShort *arr = (HArrayOfShort *)obj;
 		if (elem < 0 || elem >= obj_length(arr))
 			SignalError("java.lang.ArrayIndexOutOfBoundsException", "");
 		
 		return execute_java_constructor(0,javaLangShortClass,"(S)V",
-						(jint)(unhand(arr)->body[elem]));
+						(jint)(unhand_array(arr)->body[elem]));
 	}
-	else if ( clazz == longClass ) {
+	else if ( clazz == _Jv_longClass ) {
 		HArrayOfLong *arr = (HArrayOfLong *)obj;
 		if (elem < 0 || elem >= obj_length(arr))
 			SignalError("java.lang.ArrayIndexOutOfBoundsException", "");
 		
 		return execute_java_constructor(0,javaLangLongClass,"(J)V",
-						(jlong)(unhand(arr)->body[elem]));
+						(jlong)(unhand_array(arr)->body[elem]));
 	}
-	else if ( clazz == byteClass ) {
+	else if ( clazz == _Jv_byteClass ) {
 		HArrayOfByte *arr = (HArrayOfByte *)obj;
 		if (elem < 0 || elem >= obj_length(arr))
 			SignalError("java.lang.ArrayIndexOutOfBoundsException", "");
 		
 		return execute_java_constructor(0,javaLangByteClass,"(B)V",
-						(jint)(unhand(arr)->body[elem]));
+						(jint)(unhand_array(arr)->body[elem]));
 	}
-	else if ( clazz == booleanClass ) {
+	else if ( clazz == _Jv_booleanClass ) {
 		HArrayOfBoolean *arr = (HArrayOfBoolean *)obj;
 		if (elem < 0 || elem >= obj_length(arr))
 			SignalError("java.lang.ArrayIndexOutOfBoundsException", "");
 		
 		return execute_java_constructor(0,javaLangBooleanClass,"(Z)V",
-						(jint)(unhand(arr)->body[elem]));
+						(jint)(unhand_array(arr)->body[elem]));
 	}
-	else if ( clazz == charClass ) {
+	else if ( clazz == _Jv_charClass ) {
 		HArrayOfChar *arr = (HArrayOfChar *)obj;
 		if (elem < 0 || elem >= obj_length(arr))
 			SignalError("java.lang.ArrayIndexOutOfBoundsException", "");
 		
 		return execute_java_constructor(0,javaLangCharacterClass,"(C)V",
-						(jint)(unhand(arr)->body[elem]));
+						(jint)(unhand_array(arr)->body[elem]));
 	}
-	else if ( clazz == floatClass ) {
+	else if ( clazz == _Jv_floatClass ) {
 		HArrayOfFloat *arr = (HArrayOfFloat *)obj;
 		if (elem < 0 || elem >= obj_length(arr))
 			SignalError("java.lang.ArrayIndexOutOfBoundsException", "");
 		
 		return execute_java_constructor(0,javaLangFloatClass,"(F)V",
-						(jfloat)(unhand(arr)->body[elem]));
+						(jfloat)(unhand_array(arr)->body[elem]));
 	}
-	else if ( clazz == doubleClass ) {
+	else if ( clazz == _Jv_doubleClass ) {
 		HArrayOfDouble *arr = (HArrayOfDouble *)obj;
 		if (elem < 0 || elem >= obj_length(arr))
 			SignalError("java.lang.ArrayIndexOutOfBoundsException", "");
 		
 		return execute_java_constructor(0,javaLangDoubleClass,"(D)V",
-						(jdouble)(unhand(arr)->body[elem]));
+						(jdouble)(unhand_array(arr)->body[elem]));
 	}
 	else {
 		/* If clazz isn't one of the above then it's either a
@@ -119,7 +121,7 @@ java_lang_reflect_Array_get(struct Hjava_lang_Object* obj, jint elem)
 		if (elem < 0 || elem >= obj_length(arr))
 			SignalError("java.lang.ArrayIndexOutOfBoundsException", "");
 		
-		return unhand(arr)->body[elem];
+		return unhand_array(arr)->body[elem];
 	}
 }
 
@@ -133,12 +135,12 @@ java_lang_reflect_Array_getBoolean(struct Hjava_lang_Object* obj, jint elem)
 
 	clazz = CLASS_ELEMENT_TYPE(OBJECT_CLASS(obj));
 	
-	if ( clazz == booleanClass ) {
+	if ( clazz == _Jv_booleanClass ) {
 		HArrayOfBoolean *arr = (HArrayOfBoolean *)obj;
 		if (elem < 0 || elem >= obj_length(arr))
 			SignalError("java.lang.ArrayIndexOutOfBoundsException", "");
 		
-		return unhand(arr)->body[elem];
+		return unhand_array(arr)->body[elem];
 	} else {
 		SignalError("java.lang.IllegalArgumentException", "");
 	}
@@ -154,12 +156,12 @@ java_lang_reflect_Array_getByte(struct Hjava_lang_Object* obj, jint elem)
 
 	clazz = CLASS_ELEMENT_TYPE(OBJECT_CLASS(obj));
 	
-	if ( clazz == byteClass ) {
+	if ( clazz == _Jv_byteClass ) {
 		HArrayOfByte *arr = (HArrayOfByte *)obj;
 		if (elem < 0 || elem >= obj_length(arr))
 			SignalError("java.lang.ArrayIndexOutOfBoundsException", "");
 		
-		return unhand(arr)->body[elem];
+		return unhand_array(arr)->body[elem];
 	} else {
 		SignalError("java.lang.IllegalArgumentException", "");
 	}
@@ -175,18 +177,18 @@ java_lang_reflect_Array_getChar(struct Hjava_lang_Object* obj, jint elem)
 
 	clazz = CLASS_ELEMENT_TYPE(OBJECT_CLASS(obj));
 	
-	if ( clazz == charClass ) {
+	if ( clazz == _Jv_charClass ) {
 		HArrayOfChar *arr = (HArrayOfChar *)obj;
 		if (elem < 0 || elem >= obj_length(arr))
 			SignalError("java.lang.ArrayIndexOutOfBoundsException", "");
 		
-		return unhand(arr)->body[elem];
-	} else if ( clazz == byteClass ) {
+		return unhand_array(arr)->body[elem];
+	} else if ( clazz == _Jv_byteClass ) {
 		HArrayOfByte *arr = (HArrayOfByte *)obj;
 		if (elem < 0 || elem >= obj_length(arr))
 			SignalError("java.lang.ArrayIndexOutOfBoundsException", "");
 		
-		return unhand(arr)->body[elem];
+		return unhand_array(arr)->body[elem];
 	} else {
 		SignalError("java.lang.IllegalArgumentException", "");
 	}
@@ -202,18 +204,18 @@ java_lang_reflect_Array_getShort(struct Hjava_lang_Object* obj, jint elem)
 
 	clazz = CLASS_ELEMENT_TYPE(OBJECT_CLASS(obj));
 	
-	if ( clazz == shortClass ) {
+	if ( clazz == _Jv_shortClass ) {
 		HArrayOfShort *arr = (HArrayOfShort *)obj;
 		if (elem < 0 || elem >= obj_length(arr))
 			SignalError("java.lang.ArrayIndexOutOfBoundsException", "");
 		
-		return unhand(arr)->body[elem];
-	} else if ( clazz == byteClass ) {
+		return unhand_array(arr)->body[elem];
+	} else if ( clazz == _Jv_byteClass ) {
 		HArrayOfByte *arr = (HArrayOfByte *)obj;
 		if (elem < 0 || elem >= obj_length(arr))
 			SignalError("java.lang.ArrayIndexOutOfBoundsException", "");
 		
-		return unhand(arr)->body[elem];
+		return unhand_array(arr)->body[elem];
 	} else {
 		SignalError("java.lang.IllegalArgumentException", "");
 	}
@@ -229,30 +231,30 @@ java_lang_reflect_Array_getInt(struct Hjava_lang_Object* obj, jint elem)
 
 	clazz = CLASS_ELEMENT_TYPE(OBJECT_CLASS(obj));
 	
-	if ( clazz == intClass ) {
+	if ( clazz == _Jv_intClass ) {
 		HArrayOfInt *arr = (HArrayOfInt *)obj;
 		if (elem < 0 || elem >= obj_length(arr))
 			SignalError("java.lang.ArrayIndexOutOfBoundsException", "");
 		
-		return unhand(arr)->body[elem];
-	} else if ( clazz == shortClass ) {
+		return unhand_array(arr)->body[elem];
+	} else if ( clazz == _Jv_shortClass ) {
 		HArrayOfShort *arr = (HArrayOfShort *)obj;
 		if (elem < 0 || elem >= obj_length(arr))
 			SignalError("java.lang.ArrayIndexOutOfBoundsException", "");
 		
-		return unhand(arr)->body[elem];
-	} else if ( clazz == charClass ) {
+		return unhand_array(arr)->body[elem];
+	} else if ( clazz == _Jv_charClass ) {
 		HArrayOfChar *arr = (HArrayOfChar *)obj;
 		if (elem < 0 || elem >= obj_length(arr))
 			SignalError("java.lang.ArrayIndexOutOfBoundsException", "");
 		
-		return unhand(arr)->body[elem];
-	} else if ( clazz == byteClass ) {
+		return unhand_array(arr)->body[elem];
+	} else if ( clazz == _Jv_byteClass ) {
 		HArrayOfByte *arr = (HArrayOfByte *)obj;
 		if (elem < 0 || elem >= obj_length(arr))
 			SignalError("java.lang.ArrayIndexOutOfBoundsException", "");
 		
-		return unhand(arr)->body[elem];
+		return unhand_array(arr)->body[elem];
 	} else {
 		SignalError("java.lang.IllegalArgumentException", "");
 	}
@@ -268,36 +270,36 @@ java_lang_reflect_Array_getLong(struct Hjava_lang_Object* obj, jint elem)
 
 	clazz = CLASS_ELEMENT_TYPE(OBJECT_CLASS(obj));
 	
-	if ( clazz == longClass ) {
+	if ( clazz == _Jv_longClass ) {
 		HArrayOfLong *arr = (HArrayOfLong *)obj;
 		if (elem < 0 || elem >= obj_length(arr))
 			SignalError("java.lang.ArrayIndexOutOfBoundsException", "");
 		
-		return unhand(arr)->body[elem];
-	} else if ( clazz == intClass ) {
+		return unhand_array(arr)->body[elem];
+	} else if ( clazz == _Jv_intClass ) {
 		HArrayOfInt *arr = (HArrayOfInt *)obj;
 		if (elem < 0 || elem >= obj_length(arr))
 			SignalError("java.lang.ArrayIndexOutOfBoundsException", "");
 		
-		return unhand(arr)->body[elem];
-	} else if ( clazz == shortClass ) {
+		return unhand_array(arr)->body[elem];
+	} else if ( clazz == _Jv_shortClass ) {
 		HArrayOfShort *arr = (HArrayOfShort *)obj;
 		if (elem < 0 || elem >= obj_length(arr))
 			SignalError("java.lang.ArrayIndexOutOfBoundsException", "");
 		
-		return unhand(arr)->body[elem];
-	} else if ( clazz == charClass ) {
+		return unhand_array(arr)->body[elem];
+	} else if ( clazz == _Jv_charClass ) {
 		HArrayOfChar *arr = (HArrayOfChar *)obj;
 		if (elem < 0 || elem >= obj_length(arr))
 			SignalError("java.lang.ArrayIndexOutOfBoundsException", "");
 		
-		return unhand(arr)->body[elem];
-	} else if ( clazz == byteClass ) {
+		return unhand_array(arr)->body[elem];
+	} else if ( clazz == _Jv_byteClass ) {
 		HArrayOfByte *arr = (HArrayOfByte *)obj;
 		if (elem < 0 || elem >= obj_length(arr))
 			SignalError("java.lang.ArrayIndexOutOfBoundsException", "");
 		
-		return unhand(arr)->body[elem];
+		return unhand_array(arr)->body[elem];
 	} else {
 		SignalError("java.lang.IllegalArgumentException", "");
 	}
@@ -313,42 +315,42 @@ java_lang_reflect_Array_getFloat(struct Hjava_lang_Object* obj, jint elem)
 
 	clazz = CLASS_ELEMENT_TYPE(OBJECT_CLASS(obj));
 	
-	if ( clazz == floatClass ) {
+	if ( clazz == _Jv_floatClass ) {
 		HArrayOfFloat *arr = (HArrayOfFloat *)obj;
 		if (elem < 0 || elem >= obj_length(arr))
 			SignalError("java.lang.ArrayIndexOutOfBoundsException", "");
 		
-		return unhand(arr)->body[elem];
-	} else if ( clazz == longClass ) {
+		return unhand_array(arr)->body[elem];
+	} else if ( clazz == _Jv_longClass ) {
 		HArrayOfLong *arr = (HArrayOfLong *)obj;
 		if (elem < 0 || elem >= obj_length(arr))
 			SignalError("java.lang.ArrayIndexOutOfBoundsException", "");
 		
-		return unhand(arr)->body[elem];
-	} else if ( clazz == intClass ) {
+		return unhand_array(arr)->body[elem];
+	} else if ( clazz == _Jv_intClass ) {
 		HArrayOfInt *arr = (HArrayOfInt *)obj;
 		if (elem < 0 || elem >= obj_length(arr))
 			SignalError("java.lang.ArrayIndexOutOfBoundsException", "");
 		
-		return unhand(arr)->body[elem];
-	} else if ( clazz == shortClass ) {
+		return unhand_array(arr)->body[elem];
+	} else if ( clazz == _Jv_shortClass ) {
 		HArrayOfShort *arr = (HArrayOfShort *)obj;
 		if (elem < 0 || elem >= obj_length(arr))
 			SignalError("java.lang.ArrayIndexOutOfBoundsException", "");
 		
-		return unhand(arr)->body[elem];
-	} else if ( clazz == charClass ) {
+		return unhand_array(arr)->body[elem];
+	} else if ( clazz == _Jv_charClass ) {
 		HArrayOfChar *arr = (HArrayOfChar *)obj;
 		if (elem < 0 || elem >= obj_length(arr))
 			SignalError("java.lang.ArrayIndexOutOfBoundsException", "");
 		
-		return unhand(arr)->body[elem];
-	} else if ( clazz == byteClass ) {
+		return unhand_array(arr)->body[elem];
+	} else if ( clazz == _Jv_byteClass ) {
 		HArrayOfByte *arr = (HArrayOfByte *)obj;
 		if (elem < 0 || elem >= obj_length(arr))
 			SignalError("java.lang.ArrayIndexOutOfBoundsException", "");
 		
-		return unhand(arr)->body[elem];
+		return unhand_array(arr)->body[elem];
 	} else {
 		SignalError("java.lang.IllegalArgumentException", "");
 	}
@@ -364,48 +366,48 @@ java_lang_reflect_Array_getDouble(struct Hjava_lang_Object* obj, jint elem)
 
 	clazz = CLASS_ELEMENT_TYPE(OBJECT_CLASS(obj));
 	
-	if ( clazz == doubleClass ) {
+	if ( clazz == _Jv_doubleClass ) {
 		HArrayOfDouble *arr = (HArrayOfDouble *)obj;
 		if (elem < 0 || elem >= obj_length(arr))
 			SignalError("java.lang.ArrayIndexOutOfBoundsException", "");
 		
-		return unhand(arr)->body[elem];
-	} else if ( clazz == floatClass ) {
+		return unhand_array(arr)->body[elem];
+	} else if ( clazz == _Jv_floatClass ) {
 		HArrayOfFloat *arr = (HArrayOfFloat *)obj;
 		if (elem < 0 || elem >= obj_length(arr))
 			SignalError("java.lang.ArrayIndexOutOfBoundsException", "");
 		
-		return unhand(arr)->body[elem];
-	} else if ( clazz == longClass ) {
+		return unhand_array(arr)->body[elem];
+	} else if ( clazz == _Jv_longClass ) {
 		HArrayOfLong *arr = (HArrayOfLong *)obj;
 		if (elem < 0 || elem >= obj_length(arr))
 			SignalError("java.lang.ArrayIndexOutOfBoundsException", "");
 		
-		return unhand(arr)->body[elem];
-	} else if ( clazz == intClass ) {
+		return unhand_array(arr)->body[elem];
+	} else if ( clazz == _Jv_intClass ) {
 		HArrayOfInt *arr = (HArrayOfInt *)obj;
 		if (elem < 0 || elem >= obj_length(arr))
 			SignalError("java.lang.ArrayIndexOutOfBoundsException", "");
 		
-		return unhand(arr)->body[elem];
-	} else if ( clazz == shortClass ) {
+		return unhand_array(arr)->body[elem];
+	} else if ( clazz == _Jv_shortClass ) {
 		HArrayOfShort *arr = (HArrayOfShort *)obj;
 		if (elem < 0 || elem >= obj_length(arr))
 			SignalError("java.lang.ArrayIndexOutOfBoundsException", "");
 		
-		return unhand(arr)->body[elem];
-	} else if ( clazz == charClass ) {
+		return unhand_array(arr)->body[elem];
+	} else if ( clazz == _Jv_charClass ) {
 		HArrayOfChar *arr = (HArrayOfChar *)obj;
 		if (elem < 0 || elem >= obj_length(arr))
 			SignalError("java.lang.ArrayIndexOutOfBoundsException", "");
 		
-		return unhand(arr)->body[elem];
-	} else if ( clazz == byteClass ) {
+		return unhand_array(arr)->body[elem];
+	} else if ( clazz == _Jv_byteClass ) {
 		HArrayOfByte *arr = (HArrayOfByte *)obj;
 		if (elem < 0 || elem >= obj_length(arr))
 			SignalError("java.lang.ArrayIndexOutOfBoundsException", "");
 		
-		return unhand(arr)->body[elem];
+		return unhand_array(arr)->body[elem];
 	} else {
 		SignalError("java.lang.IllegalArgumentException", "");
 	}
@@ -425,7 +427,7 @@ java_lang_reflect_Array_set(struct Hjava_lang_Object* obj, jint elem, struct Hja
 			SignalError("java.lang.ArrayIndexOutOfBoundsException", "");
 		
 		if (val == NULL || soft_instanceof(CLASS_ELEMENT_TYPE(OBJECT_CLASS(obj)), val)) {
-			unhand(arr)->body[elem] = val;
+			unhand_array(arr)->body[elem] = val;
 		} else {
 			SignalError("java.lang.IllegalArgumentException", "");
 		}
@@ -472,12 +474,12 @@ java_lang_reflect_Array_setBoolean(struct Hjava_lang_Object* obj, jint elem, jbo
 
 	clazz = CLASS_ELEMENT_TYPE(OBJECT_CLASS(obj));
 	
-	if ( clazz == booleanClass ) {
+	if ( clazz == _Jv_booleanClass ) {
 		HArrayOfBoolean *arr = (HArrayOfBoolean *)obj;
 		if (elem < 0 || elem >= obj_length(arr))
 			SignalError("java.lang.ArrayIndexOutOfBoundsException", "");
 		
-		unhand(arr)->body[elem] = val;
+		unhand_array(arr)->body[elem] = val;
 	} else {
 		SignalError("java.lang.IllegalArgumentException", "");
 	}
@@ -493,48 +495,48 @@ java_lang_reflect_Array_setByte(struct Hjava_lang_Object* obj, jint elem, jbyte
 
 	clazz = CLASS_ELEMENT_TYPE(OBJECT_CLASS(obj));
 	
-	if ( clazz == byteClass ) {
+	if ( clazz == _Jv_byteClass ) {
 		HArrayOfByte *arr = (HArrayOfByte *)obj;
 		if (elem < 0 || elem >= obj_length(arr))
 			SignalError("java.lang.ArrayIndexOutOfBoundsException", "");
 		
-		unhand(arr)->body[elem] = val;
-	} else if ( clazz == shortClass ) {
+		unhand_array(arr)->body[elem] = val;
+	} else if ( clazz == _Jv_shortClass ) {
 		HArrayOfShort *arr = (HArrayOfShort *)obj;
 		if (elem < 0 || elem >= obj_length(arr))
 			SignalError("java.lang.ArrayIndexOutOfBoundsException", "");
 		
-		unhand(arr)->body[elem] = val;
-	} else if ( clazz == charClass ) {
+		unhand_array(arr)->body[elem] = val;
+	} else if ( clazz == _Jv_charClass ) {
 		HArrayOfChar *arr = (HArrayOfChar *)obj;
 		if (elem < 0 || elem >= obj_length(arr))
 			SignalError("java.lang.ArrayIndexOutOfBoundsException", "");
 		
-		unhand(arr)->body[elem] = val;
-	} else if ( clazz == intClass ) {
+		unhand_array(arr)->body[elem] = val;
+	} else if ( clazz == _Jv_intClass ) {
 		HArrayOfInt *arr = (HArrayOfInt *)obj;
 		if (elem < 0 || elem >= obj_length(arr))
 			SignalError("java.lang.ArrayIndexOutOfBoundsException", "");
 		
-		unhand(arr)->body[elem] = val;
-	} else if ( clazz == longClass ) {
+		unhand_array(arr)->body[elem] = val;
+	} else if ( clazz == _Jv_longClass ) {
 		HArrayOfLong *arr = (HArrayOfLong *)obj;
 		if (elem < 0 || elem >= obj_length(arr))
 			SignalError("java.lang.ArrayIndexOutOfBoundsException", "");
 		
-		unhand(arr)->body[elem] = val;
-	} else if ( clazz == floatClass ) {
+		unhand_array(arr)->body[elem] = val;
+	} else if ( clazz == _Jv_floatClass ) {
 		HArrayOfFloat *arr = (HArrayOfFloat *)obj;
 		if (elem < 0 || elem >= obj_length(arr))
 			SignalError("java.lang.ArrayIndexOutOfBoundsException", "");
 		
-		unhand(arr)->body[elem] = val;
-	} else if ( clazz == doubleClass ) {
+		unhand_array(arr)->body[elem] = val;
+	} else if ( clazz == _Jv_doubleClass ) {
 		HArrayOfDouble *arr = (HArrayOfDouble *)obj;
 		if (elem < 0 || elem >= obj_length(arr))
 			SignalError("java.lang.ArrayIndexOutOfBoundsException", "");
 		
-		unhand(arr)->body[elem] = val;
+		unhand_array(arr)->body[elem] = val;
 	} else {
 		SignalError("java.lang.IllegalArgumentException", "");
 	}
@@ -550,36 +552,36 @@ java_lang_reflect_Array_setChar(struct Hjava_lang_Object* obj, jint elem, jchar
 
 	clazz = CLASS_ELEMENT_TYPE(OBJECT_CLASS(obj));
 	
-	if ( clazz == charClass ) {
+	if ( clazz == _Jv_charClass ) {
 		HArrayOfChar *arr = (HArrayOfChar *)obj;
 		if (elem < 0 || elem >= obj_length(arr))
 			SignalError("java.lang.ArrayIndexOutOfBoundsException", "");
 		
-		unhand(arr)->body[elem] = val;
-	} else if ( clazz == intClass ) {
+		unhand_array(arr)->body[elem] = val;
+	} else if ( clazz == _Jv_intClass ) {
 		HArrayOfInt *arr = (HArrayOfInt *)obj;
 		if (elem < 0 || elem >= obj_length(arr))
 			SignalError("java.lang.ArrayIndexOutOfBoundsException", "");
 		
-		unhand(arr)->body[elem] = val;
-	} else if ( clazz == longClass ) {
+		unhand_array(arr)->body[elem] = val;
+	} else if ( clazz == _Jv_longClass ) {
 		HArrayOfLong *arr = (HArrayOfLong *)obj;
 		if (elem < 0 || elem >= obj_length(arr))
 			SignalError("java.lang.ArrayIndexOutOfBoundsException", "");
 		
-		unhand(arr)->body[elem] = val;
-	} else if ( clazz == floatClass ) {
+		unhand_array(arr)->body[elem] = val;
+	} else if ( clazz == _Jv_floatClass ) {
 		HArrayOfFloat *arr = (HArrayOfFloat *)obj;
 		if (elem < 0 || elem >= obj_length(arr))
 			SignalError("java.lang.ArrayIndexOutOfBoundsException", "");
 		
-		unhand(arr)->body[elem] = val;
-	} else if ( clazz == doubleClass ) {
+		unhand_array(arr)->body[elem] = val;
+	} else if ( clazz == _Jv_doubleClass ) {
 		HArrayOfDouble *arr = (HArrayOfDouble *)obj;
 		if (elem < 0 || elem >= obj_length(arr))
 			SignalError("java.lang.ArrayIndexOutOfBoundsException", "");
 		
-		unhand(arr)->body[elem] = val;
+		unhand_array(arr)->body[elem] = val;
 	} else {
 		SignalError("java.lang.IllegalArgumentException", "");
 	}
@@ -595,36 +597,36 @@ java_lang_reflect_Array_setShort(struct Hjava_lang_Object* obj, jint elem, jshor
 
 	clazz = CLASS_ELEMENT_TYPE(OBJECT_CLASS(obj));
 	
-	if ( clazz == shortClass ) {
+	if ( clazz == _Jv_shortClass ) {
 		HArrayOfShort *arr = (HArrayOfShort *)obj;
 		if (elem < 0 || elem >= obj_length(arr))
 			SignalError("java.lang.ArrayIndexOutOfBoundsException", "");
 		
-		unhand(arr)->body[elem] = val;
-	} else if ( clazz == intClass ) {
+		unhand_array(arr)->body[elem] = val;
+	} else if ( clazz == _Jv_intClass ) {
 		HArrayOfInt *arr = (HArrayOfInt *)obj;
 		if (elem < 0 || elem >= obj_length(arr))
 			SignalError("java.lang.ArrayIndexOutOfBoundsException", "");
 		
-		unhand(arr)->body[elem] = val;
-	} else if ( clazz == longClass ) {
+		unhand_array(arr)->body[elem] = val;
+	} else if ( clazz == _Jv_longClass ) {
 		HArrayOfLong *arr = (HArrayOfLong *)obj;
 		if (elem < 0 || elem >= obj_length(arr))
 			SignalError("java.lang.ArrayIndexOutOfBoundsException", "");
 		
-		unhand(arr)->body[elem] = val;
-	} else if ( clazz == floatClass ) {
+		unhand_array(arr)->body[elem] = val;
+	} else if ( clazz == _Jv_floatClass ) {
 		HArrayOfFloat *arr = (HArrayOfFloat *)obj;
 		if (elem < 0 || elem >= obj_length(arr))
 			SignalError("java.lang.ArrayIndexOutOfBoundsException", "");
 		
-		unhand(arr)->body[elem] = val;
-	} else if ( clazz == doubleClass ) {
+		unhand_array(arr)->body[elem] = val;
+	} else if ( clazz == _Jv_doubleClass ) {
 		HArrayOfDouble *arr = (HArrayOfDouble *)obj;
 		if (elem < 0 || elem >= obj_length(arr))
 			SignalError("java.lang.ArrayIndexOutOfBoundsException", "");
 		
-		unhand(arr)->body[elem] = val;
+		unhand_array(arr)->body[elem] = val;
 	} else {
 		SignalError("java.lang.IllegalArgumentException", "");
 	}
@@ -640,30 +642,30 @@ java_lang_reflect_Array_setInt(struct Hjava_lang_Object* obj, jint elem, jint va
 
 	clazz = CLASS_ELEMENT_TYPE(OBJECT_CLASS(obj));
 	
-	if ( clazz == intClass ) {
+	if ( clazz == _Jv_intClass ) {
 		HArrayOfInt *arr = (HArrayOfInt *)obj;
 		if (elem < 0 || elem >= obj_length(arr))
 			SignalError("java.lang.ArrayIndexOutOfBoundsException", "");
 		
-		unhand(arr)->body[elem] = val;
-	} else if ( clazz == longClass ) {
+		unhand_array(arr)->body[elem] = val;
+	} else if ( clazz == _Jv_longClass ) {
 		HArrayOfLong *arr = (HArrayOfLong *)obj;
 		if (elem < 0 || elem >= obj_length(arr))
 			SignalError("java.lang.ArrayIndexOutOfBoundsException", "");
 		
-		unhand(arr)->body[elem] = val;
-	} else if ( clazz == floatClass ) {
+		unhand_array(arr)->body[elem] = val;
+	} else if ( clazz == _Jv_floatClass ) {
 		HArrayOfFloat *arr = (HArrayOfFloat *)obj;
 		if (elem < 0 || elem >= obj_length(arr))
 			SignalError("java.lang.ArrayIndexOutOfBoundsException", "");
 		
-		unhand(arr)->body[elem] = val;
-	} else if ( clazz == doubleClass ) {
+		unhand_array(arr)->body[elem] = val;
+	} else if ( clazz == _Jv_doubleClass ) {
 		HArrayOfDouble *arr = (HArrayOfDouble *)obj;
 		if (elem < 0 || elem >= obj_length(arr))
 			SignalError("java.lang.ArrayIndexOutOfBoundsException", "");
 		
-		unhand(arr)->body[elem] = val;
+		unhand_array(arr)->body[elem] = val;
 	} else {
 		SignalError("java.lang.IllegalArgumentException", "");
 	}
@@ -679,24 +681,24 @@ java_lang_reflect_Array_setLong(struct Hjava_lang_Object* obj, jint elem, jlong
 
 	clazz = CLASS_ELEMENT_TYPE(OBJECT_CLASS(obj));
 	
-	if ( clazz == longClass ) {
+	if ( clazz == _Jv_longClass ) {
 		HArrayOfLong *arr = (HArrayOfLong *)obj;
 		if (elem < 0 || elem >= obj_length(arr))
 			SignalError("java.lang.ArrayIndexOutOfBoundsException", "");
 		
-		unhand(arr)->body[elem] = val;
-	} else if ( clazz == floatClass ) {
+		unhand_array(arr)->body[elem] = val;
+	} else if ( clazz == _Jv_floatClass ) {
 		HArrayOfFloat *arr = (HArrayOfFloat *)obj;
 		if (elem < 0 || elem >= obj_length(arr))
 			SignalError("java.lang.ArrayIndexOutOfBoundsException", "");
 		
-		unhand(arr)->body[elem] = val;
-	} else if ( clazz == doubleClass ) {
+		unhand_array(arr)->body[elem] = val;
+	} else if ( clazz == _Jv_doubleClass ) {
 		HArrayOfDouble *arr = (HArrayOfDouble *)obj;
 		if (elem < 0 || elem >= obj_length(arr))
 			SignalError("java.lang.ArrayIndexOutOfBoundsException", "");
 		
-		unhand(arr)->body[elem] = val;
+		unhand_array(arr)->body[elem] = val;
 	} else {
 		SignalError("java.lang.IllegalArgumentException", "");
 	}
@@ -712,18 +714,18 @@ java_lang_reflect_Array_setFloat(struct Hjava_lang_Object* obj, jint elem, jfloa
 
 	clazz = CLASS_ELEMENT_TYPE(OBJECT_CLASS(obj));
 	
-	if ( clazz == floatClass ) {
+	if ( clazz == _Jv_floatClass ) {
 		HArrayOfFloat *arr = (HArrayOfFloat *)obj;
 		if (elem < 0 || elem >= obj_length(arr))
 			SignalError("java.lang.ArrayIndexOutOfBoundsException", "");
 		
-		unhand(arr)->body[elem] = val;
-	} else if ( clazz == doubleClass ) {
+		unhand_array(arr)->body[elem] = val;
+	} else if ( clazz == _Jv_doubleClass ) {
 		HArrayOfDouble *arr = (HArrayOfDouble *)obj;
 		if (elem < 0 || elem >= obj_length(arr))
 			SignalError("java.lang.ArrayIndexOutOfBoundsException", "");
 		
-		unhand(arr)->body[elem] = val;
+		unhand_array(arr)->body[elem] = val;
 	} else {
 		SignalError("java.lang.IllegalArgumentException", "");
 	}
@@ -739,12 +741,12 @@ java_lang_reflect_Array_setDouble(struct Hjava_lang_Object* obj, jint elem, jdou
 
 	clazz = CLASS_ELEMENT_TYPE(OBJECT_CLASS(obj));
 	
-	if ( clazz == doubleClass ) {
+	if ( clazz == _Jv_doubleClass ) {
 		HArrayOfDouble *arr = (HArrayOfDouble *)obj;
 		if (elem < 0 || elem >= obj_length(arr))
 			SignalError("java.lang.ArrayIndexOutOfBoundsException", "");
 		
-		unhand(arr)->body[elem] = val;
+		unhand_array(arr)->body[elem] = val;
 	} else {
 		SignalError("java.lang.IllegalArgumentException", "");
 	}
@@ -753,7 +755,11 @@ java_lang_reflect_Array_setDouble(struct Hjava_lang_Object* obj, jint elem, jdou
 struct Hjava_lang_Object*
 java_lang_reflect_Array_newArray(struct Hjava_lang_Class* clazz, jint size)
 {
-	return (newArray(clazz, size));
+	if (size < 0) {
+		SignalError("java.lang.NegativeArraySizeException", "");
+	} else {
+		return (newArray(clazz, size));
+	}
 }
 
 struct Hjava_lang_Object*
@@ -763,19 +769,37 @@ java_lang_reflect_Array_multiNewArray(struct Hjava_lang_Class* clazz, HArrayOfIn
 	int i;
 	int s;
 	Hjava_lang_Object* array;
+	errorInfo info;
 
 	s = obj_length(sizes);
-	dims = calloc(s+1, sizeof(int));
+
+	if (s == 0) {
+		SignalError("java.lang.IllegalArgumentException", "zero dimensions");
+	}
+
+	dims = KCALLOC(s+1, sizeof(int));
+	if (!dims) {
+		postOutOfMemory(&info);
+		throwError(&info);
+	}
 
 	/* Copy dimentions into array */
 	for( i = 0; i < s; i++ ) {
-		dims[i] = unhand(sizes)->body[i];
+		dims[i] = unhand_array(sizes)->body[i];
+		if (dims[i] < 0) {
+			SignalError("java.lang.NegativeArraySizeException", "");
+		}
+		clazz = lookupArray(clazz, &info);
+		if (clazz == 0) {
+			KFREE(dims);
+			throwError(&info);
+		}
 	}
-	dims[i] = 0;
+	dims[i] = -1;
 
 	/* Create multi-dimension array */
 	array = newMultiArray(clazz, dims);
-	free(dims);
+	KFREE(dims);
 
 	return (array);
 }
